"""Functions to manage various objects associated to RailProjects"""

from __future__ import annotations

import os
import subprocess
import urllib.request

import yaml
from rail.core.factory_mixin import RailFactoryMixin

from .algorithm_factory import ALGORITHM_TYPES, RailAlgorithmFactory
from .catalog_factory import RailCatalogFactory
from .pipeline_factory import RailPipelineFactory
from .project_file_factory import RailProjectFileFactory
from .selection_factory import RailSelectionFactory
from .subsample_factory import RailSubsampleFactory

THE_FACTORIES: list[type[RailFactoryMixin]] = [
    RailAlgorithmFactory,
    RailCatalogFactory,
    RailPipelineFactory,
    RailProjectFileFactory,
    RailSelectionFactory,
    RailSubsampleFactory,
]

YAML_HANDLERS: dict[str, type[RailFactoryMixin]] = {
    factory.yaml_tag: factory for factory in THE_FACTORIES
}


# Lift the RailAlgorithmFactory class methods

load_algorithm_yaml = RailAlgorithmFactory.load_yaml

load_algorithm_yaml_tag = RailAlgorithmFactory.load_yaml_tag

print_algorithm_contents = RailAlgorithmFactory.print_contents

clear_algorithms = RailAlgorithmFactory.clear

get_algorithm_types = RailAlgorithmFactory.get_algorithm_types

get_algorithm_holder_dict = RailAlgorithmFactory.get_algorithm_holder_dict

get_algorithms = RailAlgorithmFactory.get_algorithms

get_algorithm_names = RailAlgorithmFactory.get_algorithm_names

get_algorithm = RailAlgorithmFactory.get_algorithm

get_algorithm_class = RailAlgorithmFactory.get_algorithm_class


# Lift the RailCatalogFactory class methods

load_catalog_yaml = RailCatalogFactory.load_yaml

load_catalog_yaml_tag = RailCatalogFactory.load_yaml_tag

print_catalog_contents = RailCatalogFactory.print_contents

clear_catalogs = RailCatalogFactory.clear

get_catalog_templates = RailCatalogFactory.get_catalog_templates

get_catalog_template_names = RailCatalogFactory.get_catalog_template_names

get_catalog_instances = RailCatalogFactory.get_catalog_instances

get_catalog_instance_names = RailCatalogFactory.get_catalog_instance_names

get_catalog_template = RailCatalogFactory.get_catalog_template

get_catalog_instance = RailCatalogFactory.get_catalog_instance


# Lift the RailPipelineFactory class methods

load_pipeline_yaml = RailPipelineFactory.load_yaml

load_pipeline_yaml_tag = RailPipelineFactory.load_yaml_tag

print_pipeline_contents = RailPipelineFactory.print_contents

clear_pipelines = RailPipelineFactory.clear

get_pipeline_templates = RailPipelineFactory.get_pipeline_templates

get_pipeline_template_names = RailPipelineFactory.get_pipeline_template_names

get_pipeline_instances = RailPipelineFactory.get_pipeline_instances

get_pipeline_instance_names = RailPipelineFactory.get_pipeline_instance_names

get_pipeline_template = RailPipelineFactory.get_pipeline_template

get_pipeline_instance = RailPipelineFactory.get_pipeline_instance


# Lift the RailProjectFileFactory class methods

load_project_file_yaml = RailProjectFileFactory.load_yaml

load_project_file_yaml_tag = RailProjectFileFactory.load_yaml_tag

get_file_templates = RailProjectFileFactory.get_file_templates

get_file_template_names = RailProjectFileFactory.get_file_template_names

get_file_instances = RailProjectFileFactory.get_file_instances

get_file_instance_names = RailProjectFileFactory.get_file_instance_names

get_file_template = RailProjectFileFactory.get_file_template

get_file_instance = RailProjectFileFactory.get_file_instance


# Lift the RailSelectionFactory class methods

load_selection_yaml = RailSelectionFactory.load_yaml

load_selection_yaml_tag = RailSelectionFactory.load_yaml_tag

get_selections = RailSelectionFactory.get_selections

get_selection_names = RailSelectionFactory.get_selection_names

get_selection = RailSelectionFactory.get_selection


# Lift the RailSubsampleFactory class methods

load_subsample_yaml = RailSubsampleFactory.load_yaml

load_subsample_yaml_tag = RailSubsampleFactory.load_yaml_tag

get_subsamples = RailSubsampleFactory.get_subsamples

get_subsample_names = RailSubsampleFactory.get_subsample_names

get_subsample = RailSubsampleFactory.get_subsample


# Define a few additional functions
def clear() -> None:
    """Clean all the factories"""
    for factory_ in THE_FACTORIES:
        factory_.clear()


def print_contents() -> None:
    """Print the contents of the factories"""
    for factory_ in THE_FACTORIES:
        factory_.print_contents()
        print("----------------")
        print("")


def load_yaml(yaml_file: str) -> None:
    """Read a yaml file and load the factory accordingly

    Parameters
    ----------
    yaml_file: str
        File to read

    Notes
    -----
    See class description for yaml file syntax
    """
    clear()
    with open(os.path.expandvars(yaml_file), encoding="utf-8") as fin:
        yaml_data = yaml.safe_load(fin)

    for yaml_key, yaml_item in yaml_data.items():
        if yaml_key == RailSelectionFactory.yaml_tag:
            load_selection_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailSubsampleFactory.yaml_tag:
            load_subsample_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailProjectFileFactory.yaml_tag:
            load_project_file_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailCatalogFactory.yaml_tag:
            load_catalog_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailPipelineFactory.yaml_tag:
            load_pipeline_yaml_tag(yaml_item, yaml_file)
        elif yaml_key in ALGORITHM_TYPES:
            load_algorithm_yaml_tag(yaml_item, f"{yaml_file}#{yaml_key}")
        else:  # pragma: no cover
            good_tags = ALGORITHM_TYPES + [
                RailSelectionFactory.yaml_tag,
                RailSubsampleFactory.yaml_tag,
                RailProjectFileFactory.yaml_tag,
                RailCatalogFactory.yaml_tag,
                RailPipelineFactory.yaml_tag,
            ]
            raise KeyError(f"Yaml Tag {yaml_key} not in expected keys {good_tags}")


def write_yaml(yaml_file: str) -> None:
    """Write the current contents for the factories to a yaml file

    Parameters
    ----------
    yaml_file: str
        File to write

    Notes
    -----
    See class description for yaml file syntax
    """
    yaml_dict: dict[str, dict] = {}
    for a_factory in THE_FACTORIES:
        yaml_dict.update(**a_factory.to_yaml_dict())
    with open(os.path.expandvars(yaml_file), mode="w", encoding="utf-8") as fout:
        yaml.dump(yaml_dict, fout)


def setup_project_area() -> int:  # pragma: no cover
    """Download test files to setup a project testsing area

    Returns
    -------
    int:
       0 for success, error code otherwise

    Notes
    -----
    This will download files into 'tests/temp_data', and could take a few
    minutes.

    This will not download the files if they are already present
    """

    if not os.path.exists("tests/ci_test.tgz"):
        urllib.request.urlretrieve(
            "http://s3df.slac.stanford.edu/people/echarles/xfer/ci_test.tgz",
            "tests/ci_test.tgz",
        )
        if not os.path.exists("tests/ci_test.tgz"):
            return 1

    if not os.path.exists("test/temp_data/projects"):
        status = subprocess.run(
            ["tar", "zxvf", "tests/ci_test.tgz", "-C", "tests"], check=False
        )
        if status.returncode != 0:
            return status.returncode

    if not os.path.exists("tests/temp_data/data/ci_test_v1.1.3/9924/part-0.parquet"):
        return 2

    if not os.path.exists("tests/temp_data/data/test/ci_test_blend_baseline_100k.hdf5"):
        os.makedirs("tests/temp_data/data/test", exist_ok=True)
        urllib.request.urlretrieve(
            "http://s3df.slac.stanford.edu/people/echarles/xfer/"
            "roman_rubin_2023_maglim_25.5_baseline_100k.hdf5",
            "tests/temp_data/data/test/ci_test_blend_baseline_100k.hdf5",
        )
        if not os.path.exists(
            "tests/temp_data/data/test/ci_test_blend_baseline_100k.hdf5"
        ):
            return 3
    return 0


def setup_mininal_example_files() -> int:  # pragma: no cover
    if not os.path.exists("tests/temp_data/data/test/minimal_gold_test.hdf5"):
        os.makedirs("tests/temp_data/data/test", exist_ok=True)
        urllib.request.urlretrieve(
            "http://s3df.slac.stanford.edu/people/echarles/xfer/"
            "minimal_gold_test.hdf5",
            "tests/temp_data/data/test/minimal_gold_test.hdf5",
        )
        if not os.path.exists("tests/temp_data/data/test/minimal_gold_test.hdf5"):
            return 1
    if not os.path.exists("tests/temp_data/data/train/minimal_gold_train.hdf5"):
        os.makedirs("tests/temp_data/data/train", exist_ok=True)
        urllib.request.urlretrieve(
            "http://s3df.slac.stanford.edu/people/echarles/xfer/"
            "minimal_gold_test.hdf5",
            "tests/temp_data/data/train/minimal_gold_train.hdf5",
        )
        if not os.path.exists("tests/temp_data/data/train/minimal_gold_train.hdf5"):
            return 2
    return 0


def teardown_project_area() -> None:  # pragma: no cover
    if not os.environ.get("NO_TEARDOWN"):
        os.system("\\rm -rf tests/temp_data")
        try:
            os.unlink("tests/ci_test.tgz")
        except FileNotFoundError:
            pass



import os

import pytest

from rail.projects.subsample_factory import RailSubsample, RailSubsampleFactory


def test_load_subsample_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    RailSubsampleFactory.clear()

    RailSubsampleFactory.load_yaml("tests/ci_subsamples.yaml")
    RailSubsampleFactory.print_contents()

    # Make sure the names of the subsamples got loaded
    the_subsample_names = RailSubsampleFactory.get_subsample_names()
    assert "test_100k" in the_subsample_names

    # Make sure the subsamples got loaded
    the_dict = RailSubsampleFactory.get_subsamples()
    assert isinstance(the_dict["test_100k"], RailSubsample)

    # get a specfic subsamples
    the_subsample = RailSubsampleFactory.get_subsample("test_100k")
    assert isinstance(the_subsample, RailSubsample)

    with pytest.raises(KeyError):
        RailSubsampleFactory.get_subsample("bad")

    # Test the interactive stuff
    RailSubsampleFactory.clear()

    RailSubsampleFactory.add_subsample(the_subsample)

    check_subsample = RailSubsampleFactory.get_subsample("test_100k")
    assert isinstance(check_subsample, type(the_subsample))

    # check writing the yaml dict
    RailSubsampleFactory.write_yaml("tests/temp.yaml")
    RailSubsampleFactory.clear()
    RailSubsampleFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_subsample = RailSubsampleFactory.get_subsample("test_100k")
    assert isinstance(check_subsample, type(the_subsample))



from __future__ import annotations

import os
from typing import Any

import numpy as np
from ceci.config import StageParameter

from . import plotting_functions, utility_functions
from .dataset import RailDataset
from .dataset_holder import RailDatasetHolder
from .plot_holder import RailPlotHolder
from .plotter import RailPlotter


class RailCatTruthDataset(RailDataset):
    """Dataset to hold any array of true redshifts"""

    data_types = dict(truth=np.ndarray)


class RailCatMagnitudesDataset(RailDataset):
    """Dataset to hold any array of magntidues and list of band names"""

    data_types = dict(magnitudes=np.ndarray, bands=list)


class RailCatTruthAndMagnitudesDataset(RailCatTruthDataset, RailCatMagnitudesDataset):
    """Dataset to hold any array of magntidues and list of band names
    and true (or spec) redshifts
    """

    data_types = RailCatMagnitudesDataset.data_types.copy()
    data_types.update(**RailCatTruthDataset.data_types)


class CatPlotterTruth(RailPlotter):
    """Class to make a histogram magnitudes in each band"""

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 151, fmt="%i", msg="Number of Redshift bins"),
    )

    input_type = RailCatTruthDataset

    def _make_hist_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:

        figure = plotting_functions.plot_true_nz(truth)
        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_hist_plot(
                prefix=prefix,
                truth=truth,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict


class CatPlotterMagntidues(RailPlotter):
    """Class to make a histogram magnitudes in each band"""

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        mag_min=StageParameter(float, 18.0, fmt="%0.2f", msg="Minimum Magnitude"),
        mag_max=StageParameter(float, 28.0, fmt="%0.2f", msg="Maximum Magnitude"),
        n_magbins=StageParameter(int, 101, fmt="%i", msg="Number of magnitude bins"),
    )

    input_type = RailCatMagnitudesDataset

    def _make_hist_plots(
        self,
        prefix: str,
        magnitudes: np.ndarray,
        bands: list[str],
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:

        xbins = np.linspace(
            self.config.mag_min, self.config.mag_max, self.config.n_magbins
        )
        figure = plotting_functions.plot_feature_histograms(
            magnitudes, bands, bins=xbins
        )
        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        magnitudes: np.ndarray = kwargs["magnitudes"]
        bands: list[str] = kwargs["bands"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_hist_plots(
                prefix=prefix,
                magnitudes=magnitudes,
                bands=bands,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict


class CatPlotterMagntiduesVsTruth(RailPlotter):
    """Class to make 2D histograms of magntidue
    versus true redshift in each band
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 151, fmt="%i", msg="Number of Redshift bins"),
        mag_min=StageParameter(float, 18.0, fmt="%0.2f", msg="Minimum Magnitude"),
        mag_max=StageParameter(float, 25.0, fmt="%0.2f", msg="Maximum Magnitude"),
        n_magbins=StageParameter(int, 141, fmt="%i", msg="Number of magnitude bins"),
    )

    input_type = RailCatTruthAndMagnitudesDataset

    def _make_2d_hist_plots(
        self,
        prefix: str,
        truth: np.ndarray,
        magnitudes: np.ndarray,
        bands: list[str],
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:

        xbins = np.linspace(self.config.z_min, self.config.z_min, self.config.n_zbins)
        ybins = np.linspace(
            self.config.mag_min, self.config.mag_max, self.config.n_magbins
        )

        figure = plotting_functions.plot_feature_target_hist2d(
            magnitudes, truth, bands, bins=(xbins, ybins)
        )
        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        magnitudes: np.ndarray = kwargs["magnitudes"]
        bands: list[str] = kwargs["bands"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_2d_hist_plots(
                prefix=prefix,
                truth=truth,
                magnitudes=magnitudes,
                bands=bands,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict


class CatPlotterColorsVsTruth(RailPlotter):
    """Class to make 2D histograms of colors
    versus true redshift in each band
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 151, fmt="%i", msg="Number of Redshift bins"),
        color_min=StageParameter(float, -2.0, fmt="%0.2f", msg="Minimum color"),
        color_max=StageParameter(float, 2.0, fmt="%0.2f", msg="Maximum color"),
        n_colorbins=StageParameter(int, 161, fmt="%i", msg="Number of color bins"),
    )

    input_type = RailCatTruthAndMagnitudesDataset

    def _make_2d_hist_plots(
        self,
        prefix: str,
        truth: np.ndarray,
        magnitudes: np.ndarray,
        bands: list[str],
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:

        colors = utility_functions.adjacent_band_colors(magnitudes)
        color_names = []
        for i in range(len(bands) - 1):
            color_names.append(f"{bands[i+1]} - {bands[i]}")
        xbins = np.linspace(self.config.z_min, self.config.z_min, self.config.n_zbins)
        ybins = np.linspace(
            self.config.color_min, self.config.color_max, self.config.n_colorbins
        )
        figure = plotting_functions.plot_feature_target_hist2d(
            colors, truth, color_names, bins=(xbins, ybins)
        )
        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        magnitudes: np.ndarray = kwargs["magnitudes"]
        bands: list[str] = kwargs["bands"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_2d_hist_plots(
                prefix=prefix,
                truth=truth,
                magnitudes=magnitudes,
                bands=bands,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict



import os

import pytest

from rail.projects.file_template import RailProjectFileInstance, RailProjectFileTemplate
from rail.projects.project_file_factory import RailProjectFileFactory


def test_load_file_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    RailProjectFileFactory.clear()

    RailProjectFileFactory.load_yaml("tests/ci_project_files.yaml")
    RailProjectFileFactory.print_contents()

    # Make sure the names of the file_templates got loaded
    the_file_templates_name = RailProjectFileFactory.get_file_template_names()
    assert "test_file_100k" in the_file_templates_name

    # Make sure the file_templates got loaded
    the_dict = RailProjectFileFactory.get_file_templates()
    assert isinstance(the_dict["test_file_100k"], RailProjectFileTemplate)

    # Make sure the names of the file_instances got loaded
    the_file_instance_name = RailProjectFileFactory.get_file_instance_names()
    assert "test_file_baseline_100k_ci_test_blend" in the_file_instance_name

    # Make sure the file_instances got loaded
    the_file_instances = RailProjectFileFactory.get_file_instances()
    assert isinstance(
        the_file_instances["test_file_baseline_100k_ci_test_blend"],
        RailProjectFileInstance,
    )

    # get a specfic file_templates
    the_file_template = RailProjectFileFactory.get_file_template("test_file_100k")
    assert isinstance(the_file_template, RailProjectFileTemplate)

    # make a file_instance
    check_file_instance = the_file_template.make_file_instance(
        name="dummy",
        catalogs_dir="tests/temp_data/data",
        project="ci_test",
        selection="blend",
    )
    assert isinstance(check_file_instance, RailProjectFileInstance)
    the_path = check_file_instance.resolve()
    assert os.path.exists(the_path)
    assert check_file_instance.check_file()
    assert check_file_instance.check_file(update=True)
    assert check_file_instance.check_file()

    with pytest.raises(KeyError):
        RailProjectFileFactory.get_file_template("bad")

    # get a specfic file_instance
    the_file_instance = RailProjectFileFactory.get_file_instance(
        "test_file_baseline_100k_ci_test_blend"
    )
    assert isinstance(the_file_instance, RailProjectFileInstance)

    with pytest.raises(KeyError):
        RailProjectFileFactory.get_file_instance("bad")

    # Test the interactive stuff
    RailProjectFileFactory.clear()
    RailProjectFileFactory.add_file_template(the_file_template)
    RailProjectFileFactory.add_file_instance(the_file_instance)

    check_file_instance = RailProjectFileFactory.get_file_instance(
        "test_file_baseline_100k_ci_test_blend"
    )
    assert isinstance(check_file_instance, type(the_file_instance))

    # check writing the yaml dict
    RailProjectFileFactory.write_yaml("tests/temp.yaml")
    RailProjectFileFactory.clear()
    RailProjectFileFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_file_instance = RailProjectFileFactory.get_file_instance(
        "test_file_baseline_100k_ci_test_blend"
    )
    assert isinstance(check_file_instance, type(check_file_instance))



"""
The :py:mod:`rail.projects` package collects a set of tools to manage RAIL-based data
analysis studies.  These tools help users define common pieced to analyses,
while also quickly testing many analysis variants with slight configuration
modifications.
"""

from . import library, name_utils
from .project import RailFlavor, RailProject


__all__ = ["library", "name_utils", "RailFlavor", "RailProject"]



"""Command line interface for rail-plot"""

from .plot_commands import (
    plot_cli,
    run_command,
    inspect_command,
    extract_datasets_command,
    make_plot_groups_for_dataset_list,
    make_plot_groups_for_project,
)

__all__ = [
    "plot_cli",
    "run_command",
    "inspect_command",
    "extract_datasets_command",
    "make_plot_groups_for_dataset_list",
    "make_plot_groups_for_project",
]



from __future__ import annotations

from rail.core.factory_mixin import RailFactoryMixin

from .pipeline_holder import RailPipelineInstance, RailPipelineTemplate


class RailPipelineFactory(RailFactoryMixin):
    """Factory class to make pipelines

    Expected usage is that user will define a yaml file with the various
    datasets that they wish to use with the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Pipelines:
        - PipelineTemplate:
            name: pz:
            pipeline_class: rail.pipelines.estimation.pz_all.PzPipeline
              input_catalog_template: degraded
            output_catalog_template: degraded
            input_file_templates:
              input_train:
                flavor: baseline
                tag: train
              input_test:
                flavor: baseline
                tag: test
            kwargs:
              algorithms: ['all']

    """

    yaml_tag: str = "Pipelines"

    client_classes = [RailPipelineTemplate, RailPipelineInstance]

    _instance: RailPipelineFactory | None = None

    def __init__(self) -> None:
        """C'tor, build an empty RailDatasetFactory"""
        RailFactoryMixin.__init__(self)
        self._pipeline_templates = self.add_dict(RailPipelineTemplate)
        self._pipeline_instances = self.add_dict(RailPipelineInstance)

    @classmethod
    def get_pipeline_templates(cls) -> dict[str, RailPipelineTemplate]:
        """Return the dict of all the pipeline templates"""
        return cls.instance().pipeline_templates

    @classmethod
    def get_pipeline_template_names(cls) -> list[str]:
        """Return the names of the pipeline templates"""
        return list(cls.instance().pipeline_templates.keys())

    @classmethod
    def get_pipeline_instances(cls) -> dict[str, RailPipelineInstance]:
        """Return the dict of all the pipeline instances"""
        return cls.instance().pipeline_instances

    @classmethod
    def get_pipeline_instance_names(cls) -> list[str]:
        """Return the names of the pipeline instances lists"""
        return list(cls.instance().pipeline_instances.keys())

    @classmethod
    def get_pipeline_template(cls, name: str) -> RailPipelineTemplate:
        """Get pipeline templates by it's assigned name

        Parameters
        ----------
        name: str
            Name of the pipeline templates to return

        Returns
        -------
        RailProjectPipelineTemplate:
            pipeline templates in question
        """
        try:
            return cls.instance().pipeline_templates[name]
        except KeyError as msg:
            raise KeyError(
                f"RailPipelineTemplate named {name} not found in RailPipelineFactory "
                f"{list(cls.instance().pipeline_templates.keys())}"
            ) from msg

    @classmethod
    def get_pipeline_instance(cls, name: str) -> RailPipelineInstance:
        """Get a pipeline instance by its assigned name

        Parameters
        ----------
        name: str
            Name of the pipeline instance list to return

        Returns
        -------
        RailProjectPipelineInstance:
            pipeline instance in question
        """
        try:
            return cls.instance().pipeline_instances[name]
        except KeyError as msg:
            raise KeyError(
                f"RailPipelineInstance named {name} not found in RailPipelineInstance "
                f"{list(cls.instance().pipeline_instances.keys())}"
            ) from msg

    @classmethod
    def add_pipeline_instance(cls, pipeline_instance: RailPipelineInstance) -> None:
        """Add a particular PipelineInstance to the factory"""
        cls.instance().add_to_dict(pipeline_instance)

    @classmethod
    def add_pipeline_template(cls, pipeline_template: RailPipelineTemplate) -> None:
        """Add a particular PipelineTemplate to the factory"""
        cls.instance().add_to_dict(pipeline_template)

    @property
    def pipeline_templates(self) -> dict[str, RailPipelineTemplate]:
        """Return the dictionary of pipeline templates"""
        return self._pipeline_templates

    @property
    def pipeline_instances(self) -> dict[str, RailPipelineInstance]:
        """Return the dictionary of pipeline instances"""
        return self._pipeline_instances

    def print_instance_contents(self) -> None:
        """Print the contents of the factory"""
        print("----------------")
        print("Pipelines:")
        RailFactoryMixin.print_instance_contents(self)



from __future__ import annotations

import os
from typing import Any, cast

import numpy as np
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from ceci.config import StageParameter
from rail.core.configurable import Configurable

from .dynamic_class import DynamicClass
from .arrow_utils import parse_item, filter_dataset, inner_join_datasets


class RailSubsampler(Configurable, DynamicClass):
    """Base class for subsampling ata

    The main function in this class is:
    run(...)

    This function will take the input files and make a single output file

    """

    config_options: dict[str, StageParameter] = {}
    sub_classes: dict[str, type[DynamicClass]] = {}

    def __init__(self, **kwargs: Any) -> None:
        """C'tor

        Parameters
        ----------
        **kwargs:
            Configuration parameters for this plotter, must match
            class.config_options data members
        """
        DynamicClass.__init__(self)
        Configurable.__init__(self, **kwargs)

    def get_basename_dict(self, **kwargs: Any) -> dict[str, str]:
        """Retrun a map of the basenames of the files to pull data from

        Returns
        -------
        Map from tag to basename, used to organize input files
        """
        raise NotImplementedError()

    def run(
        self,
        input_files: dict[str, list[str]],
        output: str,
    ) -> None:
        """Subsample the data

        Parameters
        ----------
        input_files: list[str]
            Input files to subsamle

        output: str
            Path to the output file
        """
        raise NotImplementedError()


class RandomSubsampler(RailSubsampler):
    """Pick a random subsample of the data"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Subsampler Name"),
        seed=StageParameter(int, 1234, fmt="%i", msg="Random number seed"),
        num_objects=StageParameter(
            int, None, fmt="%i", required=True, msg="Number of output objects"
        ),
    )

    def get_basename_dict(self, **kwargs: Any) -> dict[str, str]:
        return dict(main=cast(str, kwargs.get("basename", "")))

    def run(
        self,
        input_files: dict[str, list[str]],
        output: str,
    ) -> None:
        assert "main" in input_files
        the_files = input_files["main"]

        dataset = ds.dataset(the_files)
        num_rows = dataset.count_rows()
        print("num rows", num_rows)
        rng = np.random.default_rng(self.config.seed)
        print("sampling", self.config.num_objects)

        size = min(self.config.num_objects, num_rows)
        indices = rng.choice(num_rows, size=size, replace=False)
        subset = dataset.take(indices)
        print("writing", output)

        output_dir = os.path.dirname(output)
        os.makedirs(output_dir, exist_ok=True)
        pq.write_table(
            subset,
            output,
        )
        print("done")


class MultiCatalogSubsampler(RailSubsampler):
    """Pick a sample for multiple catalogs"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Subsampler Name"),
        seed=StageParameter(int, 1234, fmt="%i", msg="Random number seed"),
        num_objects=StageParameter(
            int, None, fmt="%i", required=True, msg="Number of output objects"
        ),
        weight_map=StageParameter(
            str, None, fmt="%s", msg="File with selection weights"
        ),
        object_id_col=StageParameter(
            str, "object_id", fmt="%s", msg="Object Id column name"
        ),
        cuts=StageParameter(dict, None, fmt="%s", msg="Selection cuts"),
        inputs=StageParameter(dict, None, fmt="%s", msg="Input catalog detatils"),
    )

    def get_basename_dict(self, **kwargs: Any) -> dict[str, str]:
        out_dict: dict[str, str] = {}
        for key, val in self.config.inputs.items():
            assert "basename" in val
            out_dict[key] = val.get("basename")
        return out_dict

    @staticmethod
    def _get_mag_columns(input_params: dict[str, Any]) -> list[str]:
        try:
            bands = input_params["bands"]
        except KeyError:  # pragma: no cover
            raise KeyError(
                f"Input paramters does not include 'bands' {list(input_params.keys())}"
            ) from None
        try:
            mag_band_name_template = input_params["mag_band_name_template"]
        except KeyError:  # pragma: no cover
            raise KeyError(
                f"Input paramters does not include 'mag_band_name_template' {list(input_params.keys())}"
            ) from None
        try:
            mag_err_band_name_template = input_params["mag_err_band_name_template"]
        except KeyError:  # pragma: no cover
            raise KeyError(
                f"Input paramters does not include 'mag_err_band_name_template' {list(input_params.keys())}"
            ) from None
        out_list: list[str] = [
            mag_band_name_template.format(band=band_) for band_ in bands
        ]
        out_list += [mag_err_band_name_template.format(band=band_) for band_ in bands]
        return out_list

    def _sub_selection(self, key: str, file_list: list[str]) -> ds.Dataset:
        sub_selection_params = self.config.inputs[key]
        if self.config.cuts is not None:
            all_cuts = self.config.cuts.copy()
        else:
            all_cuts = []
        sub_sel_cuts = sub_selection_params.get("cuts", [])
        if sub_sel_cuts:
            all_cuts += sub_sel_cuts
        parsed_cuts = parse_item(all_cuts)
        dataset = ds.dataset(file_list)
        save_cols: list[str] = [self.config.object_id_col]
        save_cols += self._get_mag_columns(sub_selection_params).copy()
        save_cols += sub_selection_params.get("extra_cols", [])
        filtered = filter_dataset(dataset, parsed_cuts, save_cols)  # type: ignore
        return filtered

    def _merge_selection(self, selected_data: dict[str, ds.Dataset]) -> ds.Dataset:
        return inner_join_datasets(selected_data, self.config.object_id_col)

    def run(
        self,
        input_files: dict[str, list[str]],
        output: str,
    ) -> None:

        selected_data = {
            key: self._sub_selection(key, val) for key, val in input_files.items()
        }
        subset = self._merge_selection(selected_data)
        num_rows = subset.count_rows()

        print("num rows selected", num_rows)

        rng = np.random.default_rng(self.config.seed)
        print("sampling", self.config.num_objects)

        size = min(self.config.num_objects, num_rows)
        indices = rng.choice(num_rows, size=size, replace=False)
        subset = subset.take(indices)
        print("writing", output)

        output_dir = os.path.dirname(output)
        os.makedirs(output_dir, exist_ok=True)
        pq.write_table(
            subset,
            output,
        )
        print("done")



import click
from rail.cli.rail.options import EnumChoice, PartialArgument, PartialOption

from rail.projects.execution import RunMode

__all__: list[str] = [
    "RunMode",
    "args",
    "basename",
    "config_path",
    "catalog_template",
    "file_template",
    "force",
    "flavor",
    "input_dir",
    "input_file",
    "input_selection",
    "input_tag",
    "label",
    "maglim",
    "model_dir",
    "model_name",
    "model_path",
    "output_dir",
    "pdf_dir",
    "pdf_path",
    "reducer_class_name",
    "run_mode",
    "selection",
    "site",
    "splitter_class_name",
    "subsampler_class_name",
    "subsample_name",
    "test_file_template",
    "train_file_template",
    "output_dir",
    "output_file",
    "output_catalog_template",
    "truth_path",
    "seed",
]


args = PartialArgument(
    "args",
    type=str,
    nargs=-1,
)


basename = PartialOption(
    "--basename",
    type=str,
    help="Basename for catalog template resolution",
)


config_file = PartialArgument(
    "config_file",
    type=click.Path(),
)


config_path = PartialOption(
    "--config_path",
    help="Path to configuration file",
    type=click.Path(),
)


catalog_template = PartialOption(
    "--catalog-template",
    type=str,
    help="Name of the catalog template to use",
)


file_template = PartialOption(
    "--file-template",
    type=str,
    help="Name of the file template to use",
)


force = PartialOption(
    "--force",
    help="Overwrite existing ceci configuration files",
    is_flag=True,
)

flavor = PartialOption(
    "--flavor",
    help="Pipeline configuraiton flavor",
    multiple=True,
    default=["baseline"],
)


label = PartialOption(
    "--label",
    help="File label (e.g., 'test' or 'train')",
    type=str,
)


reducer_class_name = PartialOption(
    "--reducer-class-name",
    help="Class for the reducer",
    type=str,
    default="roman_rubin",
)


selection = PartialOption(
    "--selection",
    help="Data selection",
    multiple=True,
    default=["gold"],
)


site = PartialOption(
    "--site",
    help="site for slurm submission",
    type=str,
    default=None,
)


splitter_class_name = PartialOption(
    "--splitter-class-name",
    help="Class for the splitter",
    type=str,
    default="random_splitter",
)


subsampler_class_name = PartialOption(
    "--subsampler-class-name",
    help="Class for the subsampler",
    type=str,
    default="random_subsample",
)


subsample_name = PartialOption(
    "--subsample-name",
    help="Name for the subsample",
    type=str,
    default=None,
)


input_dir = PartialOption(
    "--input-dir",
    help="Input Directory",
    type=click.Path(),
)


input_file = PartialOption(
    "--input-file",
    type=click.Path(),
    help="Input file",
)


input_selection = PartialOption(
    "--input-selection",
    help="Data selection",
    multiple=True,
    default=[None],
)


input_tag = PartialOption(
    "--input-tag",
    type=str,
    default=None,
    help="Input Catalog tag",
)


maglim = PartialOption(
    "--maglim",
    help="Magnitude limit",
    type=float,
    default=25.5,
)


model_dir = PartialOption(
    "--model-dir",
    help="Path to directory with model files",
    type=click.Path(),
)


model_path = PartialOption(
    "--model-path",
    help="Path to model file",
    type=click.Path(),
)


model_name = PartialOption(
    "--model-name",
    help="Model Name",
    type=str,
)


output_catalog_template = PartialOption(
    "--output-catalog-template",
    type=str,
    help="Name of the catalog template to use for output",
)


pdf_dir = PartialOption(
    "--pdf-dir",
    help="Path to directory with p(z) files",
    type=click.Path(),
)


pdf_path = PartialOption(
    "--pdf-path",
    help="Path to p(z) estimate file",
    type=click.Path(),
)


run_mode = PartialOption(
    "--run-mode",
    type=EnumChoice(RunMode),
    default="bash",
    help="Mode to run script",
)

size = PartialOption(
    "--size",
    type=int,
    default=100_000,
    help="Number of objects in file",
)

test_file_template = PartialOption(
    "--test-file-template",
    type=str,
    help="Name of the file template to use for test file",
)


train_file_template = PartialOption(
    "--train-file-template",
    type=str,
    help="Name of the file template to use for train file",
)


output_dir = PartialOption(
    "--output-dir",
    type=click.Path(),
    help="Path to directory for output",
)


output_file = PartialOption(
    "--output-file",
    type=click.Path(),
    help="Output file",
)


truth_path = PartialOption(
    "--truth-path",
    help="Path to truth redshift file",
    type=click.Path(),
)

seed = PartialOption(
    "--seed",
    help="Random seed",
    type=int,
)



import os

import pytest

from rail.plotting import control
from rail.plotting.dataset_holder import RailDatasetHolder
from rail.plotting.plot_group import RailPlotGroup
from rail.plotting.plotter import RailPlotter


def test_load_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0
    control.clear()

    # Load the testing yaml file
    control.load_yaml("tests/ci_plot_groups.yaml")

    # test the mechanism to avoid repeat loading
    control.load_yaml("tests/ci_plot_groups.yaml")

    plot_groups = control.get_plot_group_dict()

    assert isinstance(
        plot_groups["zestimate_v_ztrue_test_plots"], control.RailPlotGroup
    )
    control.print_contents()
    control.print_classes()

    control.write_yaml("tests/temp.yaml")
    control.clear()
    control.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")
    plot_groups = control.get_plot_group_dict()

    assert isinstance(
        plot_groups["zestimate_v_ztrue_test_plots"], control.RailPlotGroup
    )


def test_run(setup_project_area: int) -> None:
    assert setup_project_area == 0
    control.clear()

    _out_dict = control.run("tests/ci_plot_groups.yaml", outdir="tests/temp_data/plots")
    check_path = os.path.join(
        "tests",
        "temp_data",
        "plots",
        "blend_baseline_fzboost",
        "zestimate_v_ztrue_profile.png",
    )

    plot_group = control.get_plot_group("accuracy_v_ztrue")
    assert isinstance(plot_group, RailPlotGroup)

    assert plot_group.find_plot_path("blend_baseline_all", "accuracy_v_ztrue")
    with pytest.raises(KeyError):
        plot_group.find_plot_path("bad", "bad")
    with pytest.raises(KeyError):
        plot_group.find_plot_path("blend_baseline_all", "bad")

    assert os.path.exists(check_path)

    control.clear()
    out_dict2 = control.run(
        "tests/ci_plot_groups.yaml",
        outdir="tests/temp_data/plots",
        find_only=True,
        make_html=True,
    )

    assert out_dict2
    assert os.path.exists(
        "tests/temp_data/plots/plots_accuracy_v_ztrue.html",
    )

    plot_dict = out_dict2["blend_baseline_all"]
    assert plot_dict.name == "blend_baseline_all"

    plot_holder = plot_dict.plots["accuracy_v_ztrue"]
    assert plot_holder

    assert isinstance(plot_holder.plotter, RailPlotter)
    assert isinstance(plot_holder.dataset_holder, RailDatasetHolder)



"""Functions to control plot making in the context of a RailProject"""

from __future__ import annotations

import os
from typing import Any

import yaml
from rail.core.factory_mixin import RailFactoryMixin

from .dataset_factory import RailDatasetFactory
from .dataset_holder import RailDatasetHolder
from .plot_group import RailPlotGroup
from .plot_group_factory import RailPlotGroupFactory
from .plot_holder import RailPlotDict
from .plotter import RailPlotter
from .plotter_factory import RailPlotterFactory

THE_FACTORIES: list[type[RailFactoryMixin]] = [
    RailPlotterFactory,
    RailDatasetFactory,
    RailPlotGroupFactory,
]

YAML_HANDLERS: dict[str, type[RailFactoryMixin]] = {
    factory.yaml_tag: factory for factory in THE_FACTORIES
}


# Lift the RailDatasetFactory class methods

load_dataset_yaml = RailDatasetFactory.load_yaml

load_dataset_yaml_tag = RailDatasetFactory.load_yaml_tag

print_dataset_contents = RailDatasetFactory.print_contents

get_datasets = RailDatasetFactory.get_datasets

get_dataset_names = RailDatasetFactory.get_dataset_names

get_dataset_lists = RailDatasetFactory.get_dataset_lists

get_dataset_list_names = RailDatasetFactory.get_dataset_list_names

get_dataset = RailDatasetFactory.get_dataset

get_dataset_list = RailDatasetFactory.get_dataset_list

add_dataset = RailDatasetFactory.add_dataset

add_dataset_list = RailDatasetFactory.add_dataset_list

add_project = RailDatasetFactory.add_project


# Lift the RailPlotterFactory class methods

load_plotter_yaml = RailPlotterFactory.load_yaml

load_plotter_yaml_tag = RailPlotterFactory.load_yaml_tag

print_plotter_contents = RailPlotterFactory.print_contents

get_plotters = RailPlotterFactory.get_plotters

get_plotter_names = RailPlotterFactory.get_plotter_names

get_plotter_lists = RailPlotterFactory.get_plotter_lists

get_plotter_list_names = RailPlotterFactory.get_plotter_list_names

get_plotter = RailPlotterFactory.get_plotter

get_plotter_list = RailPlotterFactory.get_plotter_list

add_plotter = RailPlotterFactory.add_plotter

add_plotter_list = RailPlotterFactory.add_plotter_list


# Lift methods from the RailPlotGroupFactory

load_plot_group_yaml = RailPlotGroupFactory.load_yaml

load_plot_group_yaml_tag = RailPlotGroupFactory.load_yaml_tag

make_plot_group_yaml_for_dataset_list = RailPlotGroupFactory.make_yaml_for_dataset_list

make_plot_group_yaml_for_project = RailPlotGroupFactory.make_yaml_for_project

print_plot_group_contents = RailPlotGroupFactory.print_contents

get_plot_group_dict = RailPlotGroupFactory.get_plot_groups

get_plot_group_names = RailPlotGroupFactory.get_plot_group_names

get_plot_group = RailPlotGroupFactory.get_plot_group

add_plot_group = RailPlotGroupFactory.add_plot_group


# Lift methods from RailDatasetHolder

print_dataset_holder_classes = RailDatasetHolder.print_classes

get_dataset_holder_class = RailDatasetHolder.get_sub_class

load_dataset_holder_class = RailDatasetHolder.load_sub_class

create_dataset_holder_from_dict = RailDatasetHolder.create_from_dict


# Lift methods from RailPlotter

print_plotter_classes = RailPlotter.print_classes

get_plotter_class = RailPlotter.get_sub_class

load_plotter_class = RailPlotter.load_sub_class

write_plots = RailPlotter.write_plots


# Lift methods from RailPlotGroup

make_plots = RailPlotGroup.make_plots


# Define a few additional functions
def clear() -> None:
    """Clean all the factories"""
    for factory_ in THE_FACTORIES:
        factory_.clear()


def print_contents() -> None:
    """Print the contents of the factories"""
    for factory_ in THE_FACTORIES:
        factory_.print_contents()
        print("----------------")
        print("")


def print_classes() -> None:
    """Print the loaded classes"""
    RailPlotter.print_classes()
    print("----------------")
    print("")
    RailDatasetHolder.print_classes()
    print("----------------")


def run(
    yaml_file: str,
    include_groups: list[str] | None = None,
    exclude_groups: list[str] | None = None,
    **kwargs: Any,
) -> dict[str, RailPlotDict]:
    """Read a yaml file an make the corresponding plots

    Parameters
    ----------
    yaml_file: str
        Top level yaml file with definitinos

    include_groups: list[str]
        PlotGroups to explicitly include
        Use `None` for all plots

    exclude_groups: list[str]
        PlotGroups to explicity exclude
        Use `None` to not exclude anything

    Keywords
    --------
    find_only: bool=False
        If true, only look for existing plots

    save_plots: bool=True
        Save plots to disk

    purge_plots: bool=True
        Remove plots from memory after saving

    outdir: str | None
        If set, prepend this to the groups output dir

    make_html: bool
        If set, make an html page to browse plots

    Returns
    -------
    dict[str, RailPlotDict]:
        Newly created plots.   If purge=True this will be empty
    """
    clear()
    out_dict: dict[str, RailPlotDict] = {}
    load_yaml(yaml_file)
    yaml_file_dir = os.path.dirname(yaml_file)
    group_dict = get_plot_group_dict()

    include_groups = kwargs.pop("include_groups", None)
    exclude_groups = kwargs.pop("exclude_groups", None)
    make_html = kwargs.get("make_html", False)
    output_dir = kwargs.pop("outdir", None)
    if not output_dir:  # pragma: no cover
        output_dir = yaml_file_dir
    if include_groups is None or not include_groups:
        include_groups = list(group_dict.keys())
    if exclude_groups is None or not exclude_groups:
        exclude_groups = []
    for exclude_group_ in exclude_groups:  # pragma: no cover
        include_groups.remove(exclude_group_)

    output_pages: list[str] = []
    for group_ in include_groups:
        plot_group = group_dict[group_]
        out_dict.update(plot_group.run(outdir=output_dir, **kwargs))
        if make_html:
            output_pages.append(f"plots_{plot_group.config.name}.html")
    if make_html:
        RailPlotGroup.make_html_index(
            os.path.join(output_dir, "plot_index.html"), output_pages
        )
    return out_dict


def extract_datasets(
    config_file: str,
    dataset_holder_class: str,
    output_yaml: str,
    **kwargs: dict[str, Any],
) -> None:
    """Extract datasets into a yaml file

    Parameters
    ----------
    config_file: str
        Yaml project configuration file

    dataset_holder_class: str
        Class used to extract Datasets

    output_yaml: str
        Path to output file

    **kwargs:
        See notes for details

    Notes
    -----
    dataset_list_name: str
        Name for the resulting DatasetList

    selections: list[str]
        Selections to use

    flavors: list[str]
        Flavors to use

    split_by_flavor: bool
        Split dataset lists by flavor
    """
    extractor_cls = load_dataset_holder_class(dataset_holder_class)
    projects, datasets, dataset_lists = extractor_cls.generate_dataset_dict(
        project_file=config_file,
        **kwargs,
    )
    output_list: list[dict] = []
    output_list += [project_.to_yaml_dict() for project_ in projects]
    output_list += [dataset_.to_yaml_dict() for dataset_ in datasets]
    output_list += [dataset_list_.to_yaml_dict() for dataset_list_ in dataset_lists]

    output_data = dict(Data=output_list)
    with open(output_yaml, "w", encoding="utf-8") as fout:
        yaml.dump(output_data, fout)


def load_yaml(yaml_file: str) -> None:
    """Read a yaml file and load the factory accordingly

    Parameters
    ----------
    yaml_file: str
        File to read

    Notes
    -----
    See class description for yaml file syntax
    """
    with open(os.path.expandvars(yaml_file), encoding="utf-8") as fin:
        yaml_data = yaml.safe_load(fin)

    includes = yaml_data.pop("Includes", [])
    for include_ in includes:
        load_yaml(os.path.expandvars(include_))

    for yaml_key, yaml_item in yaml_data.items():
        if yaml_key == RailDatasetFactory.yaml_tag:
            load_dataset_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailPlotterFactory.yaml_tag:
            load_plotter_yaml_tag(yaml_item, yaml_file)
        elif yaml_key == RailPlotGroupFactory.yaml_tag:
            load_plot_group_yaml_tag(yaml_item, yaml_file)
        else:  # pragma: no cover
            good_tags = [
                RailDatasetFactory.yaml_tag,
                RailPlotterFactory.yaml_tag,
                RailPlotGroupFactory.yaml_tag,
            ]
            raise KeyError(f"Yaml Tag {yaml_key} not in expected keys {good_tags}")


def write_yaml(yaml_file: str) -> None:
    """Write the current contents for the factories to a yaml file

    Parameters
    ----------
    yaml_file: str
        File to write

    Notes
    -----
    See class description for yaml file syntax
    """
    yaml_dict: dict[str, dict] = {}
    for a_factory in THE_FACTORIES:
        yaml_dict.update(**a_factory.to_yaml_dict())
    with open(os.path.expandvars(yaml_file), mode="w", encoding="utf-8") as fout:
        yaml.dump(yaml_dict, fout)



from typing import Any

import numpy as np
from rail.utils.path_utils import find_rail_file


def make_band_names(template: str, bands: list[str]) -> list[str]:
    """Make a set of band names from template and a list of bands

    Parameters
    ----------
    template:
        Template to make the names

    bands:
        List of the bands to apply to the template

    Returns
    -------
    Names of the bands
    """
    return [template.format(band=band_) for band_ in bands]


def extract_data_to_2d_array(data: Any, column_names: list[str]) -> np.ndarray:
    """Extract a set of columns from a table to a 2D array

    Parameters
    ----------
    data:
        Input data

    column_names:
        Names of the columns to extract

    Returns
    -------
    Output 2D-Array
    """
    column_data = [data[column_] for column_ in column_names]
    return np.vstack(column_data).T


def get_band_values(
    input_data: np.ndarray,
    band_name_template: str,
    bands: list[str],
) -> np.ndarray:
    """Extract a set of columns from a table to a 2D array

    Parameters
    ----------
    input_data:
        Input data

    template:
        Template to make the names

    bands:
        List of the bands to apply to the template

    Returns
    -------
    Output 2D-Array
    """
    band_names = make_band_names(band_name_template, bands)
    values = extract_data_to_2d_array(input_data, band_names)
    return values


def fluxes_to_mags(
    flux: np.ndarray,
    zero_point: float = 31.4,
) -> np.ndarray:
    """Convert flux values to AB magnitudes.

    This function converts flux measurements to AB magnitudes using the
    standard astronomical magnitude formula: m = -2.5 * log10(flux) + zero_point.
    Negative or zero flux values result in NaN magnitudes.

    Parameters
    ----------
    flux
        Array of flux values. Must be non-negative for valid magnitude
        calculations. Values <= 0 will produce NaN in the output.
    zero_point
        Zero-point offset to apply to the magnitude calculation. This accounts
        for the flux scale and instrumental calibration. Default is 31.4.

    Returns
    -------
        Array of AB magnitudes with the same shape as the input flux array.
        Elements corresponding to flux <= 0 will be NaN.
    """
    # Handle negative and zero fluxes by replacing with NaN
    with np.errstate(divide="ignore", invalid="ignore"):
        mags = -2.5 * np.log10(flux) + zero_point

    return mags


def adjacent_band_colors(mags: np.ndarray) -> np.ndarray:
    """Return a set of colors using magnitudes in adjacent bands

    I.e., u-g, g-r, r-i, i-z, z-y

    Note that there will be one less color than bands

    Parameters
    ----------
    mags:
        Input data

    Returns
    -------
    Output colors
    """
    n_bands = mags.shape[-1]
    colors = [mags[:, i] - mags[:, i + 1] for i in range(n_bands - 1)]
    return np.vstack(colors).T


def build_template_dict(
    seds: list[str], filters: list[str]
) -> dict[str, tuple[np.ndarray, np.ndarray, np.ndarray, int]]:  # pragma: no cover
    """Extract AB templates

    Parameters
    ----------
    seds:
        Names of the seds to use

    filters:
        Name of the filters to use

    Returns
    -------
    Dict mapping sed name to a tuple with
    redshifts (N), mags (N, N_filter), colors (N, N_filter-1), sed_index
    """
    template_dict = {}
    for ised, sed in enumerate(seds):
        mag_data_list = []
        for filter_ in filters:
            path = find_rail_file(
                f"examples_data/estimation_data/data/AB/{sed}.{filter_}.AB"
            )
            data = np.loadtxt(path)
            _redshifts = data[:, 0]
            mags = fluxes_to_mags(data[:, 1], 31.4)
            mag_data_list.append(mags)
        mag_data = np.vstack(mag_data_list).T
        color_data = adjacent_band_colors(mag_data).T
        template_dict[sed] = (_redshifts, mag_data, color_data, ised)
    return template_dict



"""A set of utility functions to extract data for plotting from rail files"""

from __future__ import annotations

from typing import Any

import numpy as np
import qp
import tables_io
from rail.utils.catalog_utils import CatalogConfigBase

from rail.projects import RailProject, path_funcs

from . import utility_functions


def extract_z_true(
    filepath: str,
    colname: str = "redshift",
) -> np.ndarray:
    """Extract the true redshifts from a file

    Parameters
    ----------
    filepath: str
        Path to file with tabular data

    colname: str
        Name of the column with redshfits ['redshift']

    Returns
    -------
    redshifts: np.ndarray
        Redshifts in question

    Notes
    -----
    This assumes the redshifts are in a file that can be read by tables_io
    """
    truth_table = tables_io.read(filepath)
    return truth_table[colname]


def extract_z_point(
    filepath: str,
    colname: str = "zmode",
) -> np.ndarray:
    """Extract the point estimates of redshifts from a file

    Parameters
    ----------
    filepath: str
        Path to file with tabular data

    colname: str
        Name of the column with point estimates ['zmode']

    Returns
    -------
    z_estimates: np.ndarray
        Redshift estimates in question

    Notes
    -----
    This assumes the point estimates are in a qp file
    """
    qp_ens = qp.read(filepath)
    z_estimates = np.squeeze(qp_ens.ancil[colname])
    return z_estimates


def extract_mag(
    filepath: str,
    colname: str = "LSST_obs_i",
) -> np.ndarray:
    """Extract the i-mag from a file

    Parameters
    ----------
    filepath: str
        Path to file with tabular data

    colname: str
        Name of the column with redshfits ['redshift']

    Returns
    -------
    magnitude: np.ndarray
        Magnitude in question

    Notes
    -----
    This assumes the magnitude are in a file that can be read by tables_io
    """
    magnitude_table = tables_io.read(filepath)
    return magnitude_table[colname]


def extract_magnitudes(
    filepath: str,
    template: str,
    bands: list[str],
) -> np.ndarray:
    """Extract the magntidues from a file

    Parameters
    ----------
    filepath: str
        Path to file with tabular data

    template:
        Template to make the names

    bands:
        List of the bands to apply to the template

    Returns
    -------
    magnitudes: np.ndarray
        Magnitudes in question

    Notes
    -----
    This assumes the magnitude are in a file that can be read by tables_io
    """
    magnitude_table = tables_io.read(filepath)
    magnitudes = utility_functions.get_band_values(magnitude_table, template, bands)
    return magnitudes


def extract_z_pdf(
    filepath: str,
) -> qp.ensemble:
    """Extract the pdf estimates of redshifts from a file

    Parameters
    ----------
    filepath: str
        Path to file with tabular data

    Returns
    -------
    z_pdf: qp.ensemble
        Redshift pdf in question

    Notes
    -----
    This assumes the point estimates are in a qp file
    """
    z_pdf = qp.read(filepath)
    return z_pdf


def extract_multiple_z_point(
    filepaths: dict[str, str],
    colname: str = "zmode",
) -> dict[str, np.ndarray]:  # pragma: no cover
    """Extract the point estimates of redshifts from several files

    Parameters
    ----------
    filepaths: dict[str, str]
        Path to file with tabular data, keys will be associatd with the various
        extracted point estimates

    colname: str
        Name of the column with point estimates ['zmode']

    Returns
    -------
    z_estimates: dict[str, np.ndarray]
        Redshift estimates in question, key by the key from input argument

    Notes
    -----
    This assumes the point estimates are in a qp file
    """
    ret_dict = {key: extract_z_point(val, colname) for key, val in filepaths.items()}
    return ret_dict


def make_z_true_z_point_dict(
    z_true: np.ndarray,
    z_estimate: np.ndarray,
    mags: np.ndarray,
) -> dict[str, np.ndarray]:
    """Build a dictionary with true redshifts and a point_estimates

    Parameters
    ----------
    z_true:
        True Redshifts

    z_estimate:
        Point estimates

    mags:
        Magnitdues

    Returns
    -------
    out_dict: dict[str, np.ndarray]
        Dictionary with true redshift and a point estimate of the redshift
    """
    out_dict: dict[str, Any] = dict(
        truth=z_true,
        pointEstimate=z_estimate,
        magnitude=mags,
    )
    return out_dict


def make_z_true_multi_z_point_dict(
    z_true: np.ndarray,
    z_estimates: dict[str, np.ndarray],
) -> dict[str, Any]:
    """Build a single dictionary with true redshifts and several point_estimates

    Parameters
    ----------
    z_true: np.ndarray
        True Redshifts

    z_estimates: dict[str, np.ndarray]
        Point estimates

    Returns
    -------
    out_dict: dict[str, Any]
        Dictionary with true redshift and all the point estimate of the redshift
    """
    out_dict: dict[str, Any] = dict(
        truth=z_true,
        pointEstimates=z_estimates,
    )
    return out_dict


def get_pz_point_estimate_data(
    project: RailProject,
    selection: str,
    flavor: str,
    tag: str,
    algo: str,
) -> dict[str, np.ndarray] | None:
    """Get the true redshifts and point estimates
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz'], etc...

    tag: str
        File tag, e.g., 'test' or 'train', or 'train_zCosmos'

    Returns
    -------
    pz_data: dict[str, np.ndarray] | None
        Data in question or None if a file is missing
    """
    z_true_path = path_funcs.get_z_true_path(project, selection, flavor, tag)
    z_estimate_path = path_funcs.get_ceci_pz_output_path(
        project, selection, flavor, algo
    )
    if z_estimate_path is None:  # pragma: no cover
        return None
    z_true_data = extract_z_true(z_true_path)
    z_estimate_data = extract_z_point(z_estimate_path)
    flavor_info = project.get_flavor(flavor)
    catalog_tag = flavor_info["catalog_tag"]
    CatalogConfigBase.apply(catalog_tag)
    catalog_class = CatalogConfigBase.active_class()
    ref_band = catalog_class.band_template.format(band=catalog_class.ref_band)
    mag_data = extract_mag(z_true_path, colname=ref_band)
    pz_data = make_z_true_z_point_dict(z_true_data, z_estimate_data, mag_data)
    return pz_data


def get_ztrue_and_magntidues(
    project: RailProject,
    selection: str,
    flavor: str,
    tag: str,
) -> dict[str, np.ndarray] | None:
    """Get the true redshifts and observed magntidues
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    tag: str
        File tag, e.g., 'test' or 'train', or 'train_zCosmos'

    Returns
    -------
    out_data: dict[str, np.ndarray] | None
        Data in question or None if a file is missing
    """
    flavor_info = project.get_flavor(flavor)
    catalog_tag = flavor_info["catalog_tag"]
    CatalogConfigBase.apply(catalog_tag)
    catalog_class = CatalogConfigBase.active_class()

    z_true_path = path_funcs.get_z_true_path(project, selection, flavor, tag)
    z_true_data = extract_z_true(z_true_path, catalog_class.redshift_col)
    mag_data = extract_magnitudes(
        z_true_path, catalog_class.band_template, catalog_class.bandlist
    )
    out_data = dict(
        truth=z_true_data,
        magnitudes=mag_data,
        bands=catalog_class.bandlist,
    )
    return out_data


def get_multi_pz_point_estimate_data(
    point_estimate_infos: dict[str, dict[str, Any]],
) -> dict[str, Any] | None:
    """Get the true redshifts and point estimates

    for several analysis variants

    This checks that they all have the same redshifts
    Parameters
    ----------
    point_estimate_infos: dict[str, dict[str, Any]]
        Information about how to get point estimates

    Returns
    -------
    pz_data: dict[str, Any] | None
        Data in question or None
    """
    point_estimates: dict[str, np.ndarray] = {}
    ztrue_data: np.ndarray | None = None
    ztrue_key: str | None = None
    for key, val in point_estimate_infos.items():
        the_data = get_pz_point_estimate_data(**val)
        if the_data is None:  # pragma: no cover
            continue
        if ztrue_data is None:
            ztrue_data = the_data["truth"]
            ztrue_key = key
        else:
            if not np.allclose(ztrue_data, the_data["truth"]):  # pragma: no cover
                raise ValueError(
                    f"Mismatch in truth data. data({key}) != data({ztrue_key})"
                )
        point_estimates[key] = the_data["pointEstimate"]
    if ztrue_data is None:  # pragma: no cover
        return None
    pz_data = make_z_true_multi_z_point_dict(ztrue_data, point_estimates)
    return pz_data


def get_tomo_bins_nz_estimate_data(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str,
    classifier: str,
    summarizer: str,
) -> qp.Ensemble:
    """Get the tomographic bin n(z) estimates

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz'], etc...

    classifier: str
        Algorithm we use to make tomograpic bin

    summarizer: str
        Algorithm we use to go from p(z) to n(z)

    Returns
    -------
    nz_data: qp.Ensemble
        Tomographic bin n(z) data
    """
    paths = path_funcs.get_ceci_nz_output_paths(
        project,
        selection,
        flavor,
        algo,
        classifier,
        summarizer,
    )

    data = qp.concatenate([extract_z_pdf(path_) for path_ in paths])
    return data


def get_tomo_bins_true_nz_data(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str,
    classifier: str,
) -> qp.Ensemble:
    """Get the tomographic bin true n(z)

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz'], etc...

    classifier: str
        Algorithm we use to make tomograpic bin

    Returns
    -------
    nz_data: qp.Ensemble
        Tomographic bin n(z) data
    """
    paths = path_funcs.get_ceci_true_nz_output_paths(
        project,
        selection,
        flavor,
        algo,
        classifier,
    )

    data = qp.concatenate([extract_z_pdf(path_) for path_ in paths])
    return data



"""Functions to execute pipeline and other shell commands"""

import enum
import os
import subprocess
import time


class RunMode(enum.Enum):
    """Choose the run mode"""

    dry_run = 0
    bash = 1
    slurm = 2


S3DF_SLURM_OPTIONS: list[str] = [
    "-p",
    "milano",
    "--account",
    "rubin:commissioning@milano",
    "--mem",
    "16G",
    "--parsable",
]
PERLMUTTER_SLURM_OPTIONS: list[str] = [
    "--account",
    "m1727",
    "--constraint",
    "cpu",
    "--qos",
    "regular",
    "--parsable",
]

SLURM_OPTIONS = {
    "s3df": S3DF_SLURM_OPTIONS,
    "perlmutter": PERLMUTTER_SLURM_OPTIONS,
}


def handle_command(
    run_mode: RunMode,
    command_line: list[str],
) -> int:
    """Run a single command in the mode requested

    Parameters
    ----------
    run_mode: RunMode
        How to run the command, e.g., dry_run, bash or slurm

    command_line: list[str]
        Tokens in the command line

    Returns
    -------
    int:
        Status returned by the command.  0 for success, exit code otherwise
    """
    print("subprocess:", *command_line)
    _start_time = time.time()
    print(">>>>>>>>")
    if run_mode == RunMode.dry_run:
        # print(command_line)
        command_line.insert(0, "echo")
        finished = subprocess.run(command_line, check=False)
    elif run_mode == RunMode.bash:  # pragma: no cover
        # return os.system(command_line)
        finished = subprocess.run(command_line, check=False)
    elif run_mode == RunMode.slurm:  # pragma: no cover
        raise RuntimeError(
            "handle_command should not be called with run_mode == RunMode.slurm"
        )
    else:  # pragma: no cover
        raise AssertionError(f"Unknown run mode {run_mode}")

    returncode = finished.returncode
    _end_time = time.time()
    _elapsed_time = _end_time - _start_time
    print("<<<<<<<<")
    print(f"subprocess completed with status {returncode} in {_elapsed_time} seconds\n")
    return returncode


def handle_commands(
    run_mode: RunMode,
    command_lines: list[list[str]],
    script_path: str | None = None,
    site: str | None = None,
) -> int:  # pragma: no cover
    """Run a multiple commands in the mode requested

    Parameters
    ----------
    run_mode: RunMode
        How to run the command, e.g., dry_run, bash or slurm

    command_lines: list[list[str]]
        List of commands to run, each one is the list of tokens in the command line

    script_path: str | None
        Path to write the slurm submit script to

    Returns
    -------
    int:
        Status returned by the commands.  0 for success, exit code otherwise
    """

    if run_mode in [RunMode.dry_run, RunMode.bash]:
        for command_ in command_lines:
            retcode = handle_command(run_mode, command_)
            if retcode:
                return retcode
        return 0

    # At this point we are using slurm and need a script to send to batch
    if script_path is None:
        raise ValueError(
            "handle_commands with run_mode == RunMode.slurm requires a path to a script to write",
        )

    if site is None:
        raise ValueError(
            "handle_commands with run_mode == RunMode.slurm requires a site",
        )

    slurm_options = SLURM_OPTIONS[site]

    try:
        os.makedirs(os.path.dirname(script_path))
    except FileExistsError:
        pass
    with open(script_path, "w", encoding="utf-8") as fout:
        fout.write("#!/usr/bin/bash\n\n")
        for command_ in command_lines:
            com_line = " ".join(command_)
            fout.write(f"{com_line}\n")

    script_out = script_path.replace(".sh", ".out")

    command_line = (
        ["srun"] + slurm_options + ["--output", script_out, "--error", script_path]
    )
    try:
        with subprocess.Popen(
            command_line,
            stdout=subprocess.PIPE,
        ) as srun:
            assert srun.stdout
            line = srun.stdout.read().decode().strip()
            ret_val = int(line.split("|")[0])
    except TypeError as msg:
        raise TypeError(f"Bad slurm submit: {msg}") from msg

    return ret_val


def sbatch_wrap(
    run_mode: RunMode, site: str, args: list[str]
) -> int:  # pragma: no cover
    """Wrap a rail_pipe command with site-based arguements

    Parameters
    ----------
    run_mode: RunMode
        How to run the command, e.g., dry_run, bash or slurm

    site: str
        Execution site, used to set sbatch options

    args: list[str]
        Additional arguments

    Returns
    -------
    int
        Status.  0 for success, exit code otherwise
    """
    try:
        slurm_options = SLURM_OPTIONS[site]
    except KeyError as msg:
        raise KeyError(
            f"{site} is not a recognized site, options are {SLURM_OPTIONS.keys()}"
        ) from msg
    command_line = (
        ["sbatch"]
        + slurm_options
        + ["rail-project", "--run_mode", "slurm"]
        + list(args)
    )
    return handle_command(run_mode, command_line)



import os

import pytest

from rail.projects.selection_factory import RailSelection, RailSelectionFactory


def test_load_selection_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    RailSelectionFactory.clear()

    RailSelectionFactory.load_yaml("tests/ci_selections.yaml")
    RailSelectionFactory.print_contents()

    # Make sure the names of the selections got loaded
    the_selection_names = RailSelectionFactory.get_selection_names()
    assert "gold" in the_selection_names

    # Make sure the selections got loaded
    the_dict = RailSelectionFactory.get_selections()
    assert isinstance(the_dict["gold"], RailSelection)

    # get a specfic selections
    the_selection = RailSelectionFactory.get_selection("gold")
    assert isinstance(the_selection, RailSelection)

    with pytest.raises(KeyError):
        RailSelectionFactory.get_selection("bad")

    # Test the interactive stuff
    RailSelectionFactory.clear()

    RailSelectionFactory.add_selection(the_selection)

    check_selection = RailSelectionFactory.get_selection("gold")
    assert isinstance(check_selection, type(the_selection))

    # check writing the yaml dict
    RailSelectionFactory.write_yaml("tests/temp.yaml")
    RailSelectionFactory.clear()
    RailSelectionFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_selection = RailSelectionFactory.get_selection("gold")
    assert isinstance(check_selection, type(the_selection))



from __future__ import annotations

from types import GenericAlias
from typing import Any, get_origin


def validate_inputs(a_class: type, expected_inputs: dict, **kwargs: Any) -> None:
    """Validate that the kwargs given to a class contructor
    match the expected inputs

    Raises
    ------
    TypeError if a kwarg is not of the expected type

    KeyError is a kwaags is not in the set of expected inptus
    """
    for key, expected_type in expected_inputs.items():
        try:
            data = kwargs[key]
        except KeyError as missing_key:
            raise KeyError(
                f"{key} not provided to {a_class.__name__} in {list(kwargs.keys())}"
            ) from missing_key
        if isinstance(expected_type, GenericAlias):
            origin = get_origin(expected_type)
            if origin is not None and not isinstance(data, origin):
                raise TypeError(
                    f"{key} provided to {a_class.__name__} was "
                    f"{type(data)}, not {expected_type.__origin__}"
                )  # pragma: no cover
            continue  # pragma: no cover
        if not isinstance(data, expected_type):  # pragma: no cover
            raise TypeError(
                f"{key} provided to {a_class.__name__} was {type(data)}, expected {expected_type}"
            )



from __future__ import annotations

import os
from typing import Any

import matplotlib as mpl
import numpy as np
import qp
from ceci.config import StageParameter
from matplotlib import pyplot as plt

from .dataset import RailDataset
from .dataset_holder import RailDatasetHolder
from .plot_holder import RailPlotHolder
from .plotter import RailPlotter


class RailNZTomoBinsDataset(RailDataset):
    """Dataet to hold a n(z) distributions for a set of tomographic bins and the
    correspoding true n(z) distributions.
    """

    data_types = dict(
        truth=qp.Ensemble,
        nz_estimates=qp.Ensemble,
    )


class NZPlotterTomoBins(RailPlotter):
    """Class to make a histogram of all the nz distributions"""

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 50, fmt="%i", msg="Number of z bins"),
    )

    input_type = RailNZTomoBinsDataset

    def _make_plot(
        self,
        prefix: str,
        truth: qp.Ensemble,
        nz_estimates: qp.Ensemble,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        n_pdf = truth.npdf
        bin_edges = np.linspace(
            self.config.z_min, self.config.z_max, self.config.n_zbins + 1
        )
        truth_vals = truth.pdf(bin_edges)
        nz_vals = nz_estimates.pdf(bin_edges)

        # Compute means and variances
        z_centers = 0.5 * (bin_edges[1:] + bin_edges[:-1])
        truth_means = np.array(
            [
                np.sum(z_centers * truth_vals[i][:-1]) / np.sum(truth_vals[i][:-1])
                for i in range(n_pdf)
            ]
        )
        truth_vars = np.array(
            [
                np.sum((z_centers - truth_means[i]) ** 2 * truth_vals[i][:-1])
                / np.sum(truth_vals[i][:-1])
                for i in range(n_pdf)
            ]
        )

        est_means = np.array(
            [
                np.sum(z_centers * nz_vals[i][:-1]) / np.sum(nz_vals[i][:-1])
                for i in range(n_pdf)
            ]
        )
        est_vars = np.array(
            [
                np.sum((z_centers - est_means[i]) ** 2 * nz_vals[i][:-1])
                / np.sum(nz_vals[i][:-1])
                for i in range(n_pdf)
            ]
        )

        # Create subplots
        fig, axes = plt.subplots(n_pdf, 1, figsize=(8, 1.5 * n_pdf), sharex=True)
        if n_pdf == 1:  # pragma: no cover
            axes = [axes]  # Ensure iterable

        cmap = mpl.colormaps["plasma"]
        colors = cmap(np.linspace(0, 1, n_pdf))

        for i in range(n_pdf):
            ax = axes[i]
            color = colors[i]
            ax.plot(bin_edges, truth_vals[i], "-", color=color, label="True")
            ax.plot(bin_edges, nz_vals[i], "--", color=color, label="Estimate")

            d_mean = est_means[i] - truth_means[i]
            d_sigma = np.sqrt(est_vars[i] - truth_vars[i])

            ax.legend(
                title=f"$\\Delta \\mu_z$ = {d_mean:.3f}\n$\\Delta \\sigma_z$ = {d_sigma:.3f}",
                loc="upper right",
                fontsize="small",
                title_fontsize="small",
            )
            ax.set_ylabel(f"$n(z,{{{i}}})$")

        axes[-1].set_xlabel("z")
        fig.tight_layout()

        plot_name = self._make_full_plot_name(prefix, "")
        return RailPlotHolder(
            name=plot_name, figure=fig, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: qp.Ensemble = kwargs["truth"]
        nz_estimates: qp.Ensemble = kwargs["nz_estimates"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_plot(
                prefix=prefix,
                truth=truth,
                nz_estimates=nz_estimates,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict



import os

import pytest

from rail.plotting import pz_plotters
from rail.plotting.plotter import RailPlotter, RailPlotterList
from rail.plotting.plotter_factory import RailPlotterFactory


def test_load_yaml() -> None:
    # Load the testing yaml file
    RailPlotterFactory.clear()
    RailPlotterFactory.load_yaml("tests/ci_plots.yaml")

    # test the mechanism to avoid repeat loading
    RailPlotterFactory.load_yaml("tests/ci_plots.yaml")
    RailPlotterFactory.print_contents()
    RailPlotter.print_classes()

    # Make sure the names of the plotters got loaded
    the_names = RailPlotterFactory.get_plotter_names()
    assert "zestimate_v_ztrue_hist2d" in the_names

    # Make sure the plotters got loaded
    the_dict = RailPlotterFactory.get_plotters()
    assert isinstance(
        the_dict["zestimate_v_ztrue_hist2d"],
        pz_plotters.PZPlotterPointEstimateVsTrueHist2D,
    )

    # Make sure the names of the plotter lists got loaded
    plotter_list_names = RailPlotterFactory.get_plotter_list_names()
    assert "zestimate_v_ztrue" in plotter_list_names

    # Make sure the plotter lists got loaded
    list_dict = RailPlotterFactory.get_plotter_lists()
    assert len(list_dict["zestimate_v_ztrue"].config.plotters) == 2

    # Get a plotter by name
    a_plotter = RailPlotterFactory.get_plotter("zestimate_v_ztrue_hist2d")
    assert isinstance(a_plotter, pz_plotters.PZPlotterPointEstimateVsTrueHist2D)

    with pytest.raises(KeyError):
        RailPlotterFactory.get_plotter("bad")

    # Get a plotter list by name
    a_plotter_list = RailPlotterFactory.get_plotter_list("zestimate_v_ztrue")
    plotters = a_plotter_list.resolve(RailPlotterFactory.instance())
    assert isinstance(plotters[0], pz_plotters.PZPlotterPointEstimateVsTrueHist2D)

    with pytest.raises(KeyError):
        RailPlotterFactory.get_plotter_list("bad")

    with pytest.raises(KeyError):
        RailPlotter.get_sub_class("bad")

    # Test the interactive stuff
    RailPlotterFactory.clear()

    RailPlotterFactory.add_plotter(a_plotter)
    RailPlotterFactory.add_plotter_list(
        RailPlotterList(
            name="test_list",
            dataset_holder_class="rail.plotting.pz_data_holders.RailPZPointEstimateDataHolder",
            plotters=[a_plotter.config.name],
        )
    )

    check_plotter = RailPlotterFactory.get_plotter("zestimate_v_ztrue_hist2d")
    assert isinstance(check_plotter, type(a_plotter))

    check_list = RailPlotterFactory.get_plotter_list("test_list")
    assert a_plotter.config.name in check_list.config.plotters

    # check writing the yaml dict
    RailPlotterFactory.write_yaml("tests/temp.yaml")
    RailPlotterFactory.clear()
    RailPlotterFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_plotter = RailPlotterFactory.get_plotter("zestimate_v_ztrue_hist2d")
    assert isinstance(check_plotter, type(a_plotter))

    check_list = RailPlotterFactory.get_plotter_list("test_list")
    assert a_plotter.config.name in check_list.config.plotters



import os

import pytest
from click.testing import CliRunner, Result

from rail.cli.rail_project.project_commands import project_cli


def check_result(
    result: Result,
) -> None:
    if not result.exit_code == 0:
        raise ValueError(f"{result} failed with {result.exit_code} {result.output}")


def test_cli_help() -> None:
    runner = CliRunner()

    result = runner.invoke(project_cli, "--help")
    check_result(result)


def test_cli_inspect() -> None:
    runner = CliRunner()

    result = runner.invoke(project_cli, "inspect tests/ci_project.yaml")
    check_result(result)


def test_cli_build() -> None:
    runner = CliRunner()

    os.system("\\rm -rf tests/temp_data/projects/ci_test/pipelines")
    os.system("\\rm -rf tests/temp_data/projects/ci_test/logs")

    result = runner.invoke(project_cli, "build --flavor all tests/ci_project.yaml")
    check_result(result)


def test_cli_reduce() -> None:
    runner = CliRunner()

    result = runner.invoke(
        project_cli,
        "reduce "
        "--catalog-template truth "
        "--output-catalog-template reduced "
        "--reducer-class-name roman_rubin "
        "--selection gold "
        "tests/ci_project.yaml",
    )
    check_result(result)


def test_cli_subsample() -> None:
    runner = CliRunner()

    result = runner.invoke(
        project_cli,
        "subsample "
        "--catalog-template degraded "
        "--flavor baseline "
        "--file-template test_file_100k "
        "--subsampler-class-name random_subsampler "
        "--subsample-name test_100k "
        "--run-mode dry_run "
        "--selection gold "
        "--basename output.hdf5 "
        "tests/ci_project.yaml",
    )
    check_result(result)


def test_cli_split() -> None:
    runner = CliRunner()

    result = runner.invoke(
        project_cli,
        "split "
        "--flavor baseline "
        "--file-template train_file_10 "
        "--test-file-template test_split_file "
        "--train-file-template train_split_file "
        "--splitter-class-name random_splitter "
        "--run-mode dry_run "
        "--selection gold "
        "tests/ci_project.yaml",
    )
    check_result(result)


@pytest.mark.parametrize(
    "pipeline",
    [
        "blending",
        "estimate",
        "evaluate",
        "inform",
        "phot-errors",
        "prepare",
        "pz",
        "inform-sompz",
        "estimate-sompz",
        "inform-recalib",
        "estimate-recalib",
        "inform-somlike",
        "somlike-recalib",
        "spec-selection",
        "tomography",
        "truth-to-observed",
    ],
)
def test_cli_run(pipeline: str) -> None:
    runner = CliRunner()

    if pipeline == "estimate":
        label_str = "--input-tag test "
    else:
        label_str = ""

    result = runner.invoke(
        project_cli,
        f"run {pipeline} --selection gold --flavor baseline {label_str} "
        "--run-mode dry_run tests/ci_project.yaml",
    )
    check_result(result)


def test_cli_wrap_model(setup_project_area: int) -> None:
    assert setup_project_area == 0
    runner = CliRunner()

    result = runner.invoke(
        project_cli,
        "wrap-pz-models "
        "--outdir tests/temp_data "
        "--flavor baseline "
        "--selection blend "
        "tests/ci_project.yaml",
    )
    check_result(result)



"""
Utility code to help define standard paths for various data products
"""

import copy
import os
import re
from functools import partial
from typing import Any

CommonPaths = dict(
    root=".",
    scratch_root=".",
    project="",
    project_dir="{root}/projects/{project}",
    project_scratch_dir="{scratch_root}/projects/{project}",
    catalogs_dir="{root}/catalogs",
    pipelines_dir="{project_dir}/pipelines",
)

PathTemplates = dict(
    pipeline_path="{pipelines_dir}/{pipeline}_{flavor}.yaml",
    ceci_output_dir="{project_dir}/data/{selection}_{flavor}",
    ceci_file_path="{tag}_{stage}.{suffix}",
)


def update_include_dict(
    orig_dict: dict[str, Any],
    include_dict: dict[str, Any],
) -> None:  # pragma: no cover
    """Update a dict by updating (instead of replacing) sub-dicts

    Parameters
    ----------
    orig_dict: dict[str, Any]
        Original dict
    include_dict: dict[str, Any],
        Dict used to update the original
    """
    for key, val in include_dict.items():
        if isinstance(val, dict) and key in orig_dict:
            sub_dict = orig_dict[key]
            if isinstance(sub_dict, dict):
                update_include_dict(orig_dict[key], val)
                continue
        if isinstance(val, (dict, list)):
            orig_dict[key] = val.copy()
        else:
            orig_dict[key] = val


def get_required_interpolants(template: str) -> list[str]:
    """Get the list of interpolants required to format a template string

    Notes
    -----
    'interpolants' are strings that must be replaced in to format a string,
    e.g., in "{project_dir}/models" "{project_dir}" would an interpolant
    """
    return re.findall("{.*?}", template)


def format_template(template: str, **kwargs: Any) -> str:
    """Resolve a specific template

    This is fault-tolerant and will not raise KeyError if some
    of the required interpolants are missing, but rather just
    leave them untouched
    """

    required_interpolants = re.findall("{.*?}", template)
    interpolants = kwargs.copy()

    env_interpolants = {
        key: val
        for key, val in os.environ.items()
        if (key.find("PZ") == 0 or key.find("RAIL") == 0)
    }
    interpolants.update(**env_interpolants)

    for interpolant_ in required_interpolants:
        interpolants.setdefault(
            interpolant_.replace("}", "").replace("{", ""), interpolant_
        )
    return template.format(**interpolants)


def resolve_dict(source: dict, interpolants: dict) -> dict:
    """Recursively resolve a dictionary using interpolants

    Parameters
    ----------
    source: dict
        Dictionary of string templates

    interpolants: dict
        Dictionary of strings used to resolve templates

    Returns
    -------
    dict:
        Dictionary of resolved templates
    """
    env_interpolants = {
        key: val
        for key, val in os.environ.items()
        if (key.find("PZ") == 0 or key.find("RAIL") == 0)
    }
    interpolants.update(**env_interpolants)
    if source:
        sink = copy.deepcopy(source)
        for k, v in source.items():
            v_interpolated: list | dict | str = ""
            match v:
                case dict():
                    v_interpolated = resolve_dict(source[k], interpolants)
                # Lists are not being used yet
                case list():  # pragma: no cover
                    v_interpolated = [resolve_dict(_v, interpolants) for _v in v]
                case str():
                    v_interpolated = v.format(**interpolants)
                case _:
                    raise ValueError("Cannot interpolate type!")

            sink[k] = v_interpolated
    else:
        sink = {}

    return sink


def _resolve(
    templates: dict, source: dict, interpolants: dict
) -> dict:  # pragma: no cover
    """Resolve a set of templates using interpolants and allow for overrides

    Parameters
    ----------
    templates: dict
        Dictionary of string templates

    source: dict
        Dictionary of overrides

    interpolants: dict
        Dictionary of strings used to resolve templates


    Returns
    -------
    dict:
        Dictionary of resoluved templates
    """

    sink = copy.deepcopy(templates)
    if (overrides := source) is not None:
        for k, v in overrides.items():
            sink[k] = v
    for k, v in sink.items():
        match v:
            # Not being used yet:
            case partial():  # pragma: no cover
                sink[k] = v(**sink)
            case _:
                continue
    sink = resolve_dict(sink, interpolants)
    return sink


class NameFactory:
    """Class defining standard paths for various data products"""

    config_template = dict(
        CommonPaths=CommonPaths,
        PathTemplates=PathTemplates,
    )

    def __init__(
        self,
        config: dict | None = None,
        templates: dict | None = None,
        interpolants: dict | None = None,
    ):
        """C'tor"""
        if config is None:
            config = {}
        if templates is None:
            templates = {}
        if interpolants is None:
            interpolants = {}

        self._config = copy.deepcopy(self.config_template)
        for key, _val in config.items():
            if key in self._config:
                self._config[key].update(**config[key])
        self._templates = copy.deepcopy(self._config["PathTemplates"])
        self._templates.update(**templates)
        self._interpolants: dict = {}

        self.templates = {}
        for k, v in templates.items():
            self.templates[k] = partial(v.format, **templates)

        self.interpolants = self._config["CommonPaths"]
        self.interpolants = interpolants

    def get_path_templates(self) -> dict:
        return self._config["PathTemplates"]

    def get_common_paths(self) -> dict:
        return self._config["CommonPaths"]

    def get_resolved_common_paths(self) -> dict:
        return resolve_dict(self._config["CommonPaths"], self._interpolants)

    @property
    def interpolants(self) -> dict:
        """Return the dict of interpolants that are used to resolve templates"""
        return self._interpolants

    @interpolants.setter
    def interpolants(self, config: dict) -> None:
        """Update the dict of interpolants that are used to resolve templates"""
        env_interpolants = {
            key: val
            for key, val in os.environ.items()
            if (key.find("PZ") == 0 or key.find("RAIL") == 0)
        }

        for key, value in config.items():
            new_value = value.format(**self.interpolants, **env_interpolants)
            self.interpolants[key] = new_value

    @interpolants.deleter
    def interpolants(self) -> None:
        """Reset the dict of interpolants that are used to resolve templates"""
        self._interpolants = {}

    def resolve_from_config(self, config: dict) -> dict:  # pragma: no cover
        """Resolve all the templates in a dict

        Parameters
        ----------
        config: dict
            Dictionary containing templates to be resolved

        Returns
        -------
        dict:
            Dictionary with resolved versions of the templates
        """
        resolved = _resolve(
            self.templates,
            config,
            self.interpolants,
        )
        config.update(resolved)

        return resolved

    def resolve_path(self, config: dict, path_key: str, **kwargs: Any) -> str:
        """Resolve a particular template in a config dict

        Parameters
        ----------
        config: dict
            Dictionary containing templates to be resolved

        path_key: str
            Key for the specific template

        Returns
        -------
        str:
            Resolved version of the template
        """
        try:
            path_value = config[path_key]
            formatted = format_template(path_value, **kwargs, **self.interpolants)
            return formatted
        except KeyError as missing_key:
            raise KeyError(f"Path '{path_key}' not found in {config}") from missing_key

    def get_template(self, section_key: str, path_key: str) -> str:
        """Return the template for a particular file type

        Parameters
        ----------
        section_key: str
            Which part of the config to look in
            E.g., (CommonPaths, PathTemplates, Files)

        path_key: str
            Key for the specific template

        Returns
        -------
        str:
            Template for file of this type
        """
        try:
            section = self._config[section_key]
        except KeyError as msg:
            raise KeyError(
                f"Config section {section_key} not present:"
                f"available sections are {list(self._config.keys())}",
            ) from msg
        try:
            return section[path_key]
        except KeyError as msg:
            raise KeyError(
                f"Config key {path_key} not present in {section_key}:"
                f"available paths are {list(section.keys())}",
            ) from msg

    def resolve_template(self, section_key: str, path_key: str, **kwargs: Any) -> str:
        """Return the template for a particular file type

        Parameters
        ----------
        section_key: str
            Which part of the config to look in
            E.g., (CommonPaths, PathTemplates, Files)

        path_key: str
            Key for the specific template

        Returns
        -------
        str:
            Resolved path
        """
        template = self.get_template(section_key, path_key)
        return format_template(template, **self.interpolants, **kwargs)

    def resolve_path_template(self, path_key: str, **kwargs: Any) -> str:
        """Return a particular path templated

        Parameters
        ----------
        path_key: str
            Key for the specific template

        Returns
        -------
        str:
            Resolved path
        """
        template = self.get_template("PathTemplates", path_key)
        interp_dict = self.interpolants.copy()
        interp_dict.update(**kwargs)
        return format_template(template, **interp_dict)

    def resolve_common_path(self, path_key: str, **kwargs: Any) -> str:
        """Return a particular common path template

        Parameters
        ----------
        path_key: str
            Key for the specific template

        Returns
        -------
        str:
            Resolved path
        """
        template = self.get_template("CommonPaths", path_key)
        interp_dict = self.interpolants.copy()
        interp_dict.update(**kwargs)
        return format_template(template, **interp_dict)



"""A set of utility functions to extract data for plotting from rail files"""

from __future__ import annotations

import glob
import os

from rail.projects import RailProject


def get_z_true_path(
    project: RailProject,
    selection: str,
    flavor: str,
    tag: str,
) -> str:
    """Get the path to the the file with true redshfits
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    tag: str
        File tag, e.g., 'test' or 'train', or 'train_zCosmos'

    Returns
    -------
    path: str
        Path to the file in question
    """
    return project.get_file_for_flavor(flavor, tag, selection=selection)


def get_ceci_pz_output_path(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str,
) -> str | None:
    """Get the paths to the file with redshfit estimates
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz', etc..

    Returns
    -------
    path: str | None
        Path to the file in question, if it exists, otherwise None
    """
    outdir = project.get_path("ceci_output_dir", selection=selection, flavor=flavor)
    basename = f"output_estimate_{algo}.hdf5"
    outpath = os.path.join(outdir, basename)
    return outpath if os.path.exists(outpath) else None


def get_ceci_pz_model_paths(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str | None = None,
) -> list[str]:
    """Get the paths to the file with pz model
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project:
        Object with information about the structure of the current project

    selection:
        Data selection in question, e.g., 'gold', or 'blended'

    flavor:
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo:
        Algorithm we want the estimates for, e.g., 'knn', 'bpz', etc..
        None will find all the existing model files.

    Returns
    -------
    path: str | None
        Path to the file in question, if it exists, otherwise None
    """
    outdir = project.get_path("ceci_output_dir", selection=selection, flavor=flavor)
    if algo is None:
        basepath = os.path.join(outdir, "model_inform_*.pkl")
        paths = sorted(glob.glob(basepath))
        return [output for output in paths if os.path.exists(output)]
    basename = f"model_inform_{algo}.pkl"
    outpath = os.path.join(outdir, basename)
    return [outpath] if os.path.exists(outpath) else []


def get_ceci_nz_output_paths(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str,
    classifier: str,
    summarizer: str,
) -> list[str]:
    """Get the paths to the file with n(z) estimates
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz'], etc...

    classifier: str
        Algorithm we use to make tomograpic bin

    summarizer: str
        Algorithm we use to go from p(z) to n(z)

    Returns
    -------
    paths: list[str]
        Paths to data
    """
    outdir = project.get_path("ceci_output_dir", selection=selection, flavor=flavor)
    basename = f"single_NZ_summarize_{algo}_{classifier}_bin*_{summarizer}.hdf5"
    outpath = os.path.join(outdir, basename)
    paths = sorted(glob.glob(outpath))
    return paths


def get_ceci_true_nz_output_paths(
    project: RailProject,
    selection: str,
    flavor: str,
    algo: str,
    classifier: str,
) -> list[str]:
    """Get the paths to the file with n(z) estimates
    for a particualar analysis selection and flavor

    Parameters
    ----------
    project: RailProject
        Object with information about the structure of the current project

    selection: str
        Data selection in question, e.g., 'gold', or 'blended'

    flavor: str
        Analysis flavor in question, e.g., 'baseline' or 'zCosmos'

    algo: str
        Algorithm we want the estimates for, e.g., 'knn', 'bpz'], etc...

    classifier: str
        Algorithm we use to make tomograpic bin

    Returns
    -------
    paths: list[str]
        Paths to data
    """
    outdir = project.get_path("ceci_output_dir", selection=selection, flavor=flavor)
    basename = f"true_NZ_true_nz_{algo}_{classifier}_bin*.hdf5"
    outpath = os.path.join(outdir, basename)
    paths = sorted(glob.glob(outpath))
    return paths



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

from matplotlib.figure import Figure

if TYPE_CHECKING:
    from .dataset_holder import RailDatasetHolder
    from .plotter import RailPlotter


class RailPlotHolder:
    """Simple class for wrapping matplotlib Figure

    This includes the path to where the figure is saved
    """

    def __init__(
        self,
        name: str,
        path: str | None = None,
        figure: Figure | None = None,
        plotter: RailPlotter | None = None,
        dataset_holder: RailDatasetHolder | None = None,
    ):
        """C'tor"""
        self._name = name
        self._path = path
        self._figure = figure
        self._plotter = plotter
        self._dataset_holder = dataset_holder

    @property
    def name(self) -> str:
        """Return the name of the plot"""
        return self._name

    @property
    def path(self) -> str | None:
        """Return the path to the saved plot"""
        return self._path

    @property
    def figure(self) -> Figure | None:
        """Return the matplotlib Figure"""
        return self._figure

    @property
    def plotter(self) -> RailPlotter | None:
        """Return the object used to make the plot"""
        return self._plotter

    @property
    def dataset_holder(self) -> RailDatasetHolder | None:
        """Return the dataset used to make the plot"""
        return self._dataset_holder

    def set_path(
        self,
        path: str | None = None,
    ) -> None:
        """Set the path to the saved plot"""
        self._path = path

    def set_figure(
        self,
        figure: Figure | None = None,
    ) -> None:
        """Set the Matplotlib figure"""
        self._figure = figure

    def savefig(
        self,
        relpath: str,
        outdir: str = ".",
        **kwargs: Any,
    ) -> None:
        if self.figure is None:  # pragma: no cover
            raise ValueError(f"Tried to savefig missing a Figure {self.name}")

        self.set_path(relpath)
        fullpath = os.path.join(outdir, relpath)
        self.figure.savefig(fullpath, **kwargs)


class RailPlotDict:
    """Simple class for dicts of matplotlib Figures

    This collects a set of plots made on the same dataset.
    The key is typically the name of the RailPlotter used to make each plot.
    """

    def __init__(self, name: str, plots: dict[str, RailPlotHolder]):
        self._name = name
        self._plots = plots

    @property
    def name(self) -> str:
        """Return the name of the plot dict"""
        return self._name

    @property
    def plots(self) -> dict[str, RailPlotHolder] | None:
        """Return the dict of plots"""
        return self._plots

    def savefigs(
        self,
        outpath: str,
        figtype: str = "png",
        **kwargs: Any,
    ) -> None:
        purge = kwargs.pop("purge", False)
        if not os.path.exists(outpath):  # pragma: no cover
            os.makedirs(outpath)
        for _key, val in self._plots.items():
            if val.path:  # pragma: no cover
                val.savefig(val.path, outpath, **kwargs)
            else:
                val.savefig(
                    os.path.join(os.path.basename(outpath), f"{val.name}.{figtype}"),
                    os.path.dirname(outpath),
                    **kwargs,
                )
            if purge:
                val.set_figure(None)



from __future__ import annotations

import math
import os
from typing import Any

import numpy as np
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import yaml
from ceci.config import StageParameter
from pyarrow import acero
from rail.core.configurable import Configurable

from .arrow_utils import parse_item
from .dynamic_class import DynamicClass

COLUMNS = [
    "galaxy_id",
    "ra",
    "dec",
    "redshift",
    "LSST_obs_u",
    "LSST_obs_g",
    "LSST_obs_r",
    "LSST_obs_i",
    "LSST_obs_z",
    "LSST_obs_y",
    "ROMAN_obs_F184",
    "ROMAN_obs_J129",
    "ROMAN_obs_H158",
    "ROMAN_obs_W146",
    "ROMAN_obs_Z087",
    "ROMAN_obs_Y106",
    "ROMAN_obs_K213",
    "ROMAN_obs_R062",
    "totalEllipticity",
    "totalEllipticity1",
    "totalEllipticity2",
    "diskHalfLightRadiusArcsec",
    "spheroidHalfLightRadiusArcsec",
    "bulge_frac",
    # "healpix",
]


COLUMNS_COM_CAM = [
    "objectId",
    "coord_ra",
    "coord_dec",
    "u_cModelFlux",
    "g_cModelFlux",
    "r_cModelFlux",
    "i_cModelFlux",
    "z_cModelFlux",
    "y_cModelFlux",
    "u_cModelFluxErr",
    "g_cModelFluxErr",
    "r_cModelFluxErr",
    "i_cModelFluxErr",
    "z_cModelFluxErr",
    "y_cModelFluxErr",
]

PROJECTIONS_COM_CAM = [
    {
        "ref_flux": pc.field("i_cModelFlux"),
    }
]


PROJECTIONS_DP1 = [
    {
        "ref_flux": pc.field("i_psfFlux"),
    }
]

PROJECTIONS = [
    {
        "mag_u_lsst": pc.field("LSST_obs_u"),
        "mag_g_lsst": pc.field("LSST_obs_g"),
        "mag_r_lsst": pc.field("LSST_obs_r"),
        "mag_i_lsst": pc.field("LSST_obs_i"),
        "mag_z_lsst": pc.field("LSST_obs_z"),
        "mag_y_lsst": pc.field("LSST_obs_y"),
        "totalHalfLightRadiusArcsec": pc.add(
            pc.multiply(
                pc.field("diskHalfLightRadiusArcsec"),
                pc.subtract(pc.scalar(1), pc.field("bulge_frac")),
            ),
            pc.multiply(
                pc.field("spheroidHalfLightRadiusArcsec"),
                pc.field("bulge_frac"),
            ),
        ),
        "_orientationAngle": pc.atan2(
            pc.field("totalEllipticity2"), pc.field("totalEllipticity1")
        ),
    },
    {
        "major": pc.divide(
            pc.field("totalHalfLightRadiusArcsec"),
            pc.sqrt(pc.field("totalEllipticity")),
        ),
        "minor": pc.multiply(
            pc.field("totalHalfLightRadiusArcsec"),
            pc.sqrt(pc.field("totalEllipticity")),
        ),
        "orientationAngle": pc.multiply(
            pc.scalar(0.5),
            pc.subtract(
                pc.field("_orientationAngle"),
                pc.multiply(
                    pc.floor(
                        pc.divide(pc.field("_orientationAngle"), pc.scalar(2 * math.pi))
                    ),
                    pc.scalar(2 * math.pi),
                ),
            ),
        ),
    },
]


class RailReducer(Configurable, DynamicClass):
    """Base class for reducing data catalogs

    The main function in this class is:
    run(input_catalog, output_catalog)

    This function will files in the input_catalog, and reduce each one to make the
    output catalog
    """

    config_options: dict[str, StageParameter] = {}
    sub_classes: dict[str, type[DynamicClass]] = {}

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        **kwargs:
            Configuration parameters for this Reducer, must match
            class.config_options data members
        """
        DynamicClass.__init__(self)
        Configurable.__init__(self, **kwargs)

    def run(
        self,
        input_catalog: str,
        output_catalog: str,
    ) -> None:
        """Subsample the data

        Parameters
        ----------
        input_catalog: str,
            Input files to subsamle

        output_catalog: str,
            Path to the output file
        """
        raise NotImplementedError()


class RomanRubinReducer(RailReducer):
    """Class to reduce the 'roman_rubin' simulation input files for pz analysis"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Reducer Name"),
        cuts=StageParameter(dict, {}, fmt="%s", msg="Selections"),
    )

    def run(
        self,
        input_catalog: str,
        output_catalog: str,
    ) -> None:
        # Try to do this right
        try:
            parsed_filter = parse_item(self.config.cuts)
            predicate = pq.filters_to_expression(parsed_filter)
        except Exception as msg:
            # Fallback to old way.  FIXME, deprecate this
            if self.config.cuts:
                if "maglim_i" in self.config.cuts:
                    predicate = pc.field("LSST_obs_i") < self.config.cuts["maglim_i"][1]
                elif "maglim_Y" in self.config.cuts:
                    predicate = (
                        pc.field("ROMAN_obs_Y106") < self.config.cuts["maglim_Y"][1]
                    )
                else:
                    raise ValueError("No valid cut") from msg
            else:  # pragma: no cover
                predicate = None

        dataset = ds.dataset(
            input_catalog,
            format="parquet",
        )

        scan_node = acero.Declaration(
            "scan",
            acero.ScanNodeOptions(
                dataset,
                columns=COLUMNS,
                filter=predicate,
            ),
        )

        filter_node = acero.Declaration(
            "filter",
            acero.FilterNodeOptions(
                predicate,
            ),
        )

        column_projection = {k: pc.field(k) for k in COLUMNS}
        projection = column_projection
        project_nodes = []
        for _projection in PROJECTIONS:
            for k, v in _projection.items():
                projection[k] = v
            project_node = acero.Declaration(
                "project",
                acero.ProjectNodeOptions(
                    [v for k, v in projection.items()],
                    names=[k for k, v in projection.items()],
                ),
            )
            project_nodes.append(project_node)

        seq = [
            scan_node,
            filter_node,
            *project_nodes,
        ]
        plan = acero.Declaration.from_sequence(seq)

        # batches = plan.to_reader(use_threads=True)
        table = plan.to_table(use_threads=True)
        print(f"writing dataset to {output_catalog}")

        output_dir = os.path.dirname(output_catalog)

        os.makedirs(output_dir, exist_ok=True)
        pq.write_table(table, output_catalog)


class ComCamReducer(RailReducer):
    """Class to reduce the 'com_cam' input files for pz analysis"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Reducer Name"),
        cuts=StageParameter(dict, {}, fmt="%s", msg="Selections"),
    )

    _mag_offset = 31.4

    def run(
        self,
        input_catalog: str,
        output_catalog: str,
    ) -> None:  # pragma: no cover

        # Try to do this right
        try:
            parsed_filter = parse_item(self.config.cuts)
            predicate = pq.filters_to_expression(parsed_filter)
        except Exception:
            # Fallback to old way.  FIXME, deprecate this
            mag_cut = self.config.cuts["maglim_i"][1]
            flux_cut = np.power(10, (self._mag_offset - mag_cut) / 2.5)

            if self.config.cuts:
                predicate = pc.field("i_cModelFlux") > flux_cut
            else:  # pragma: no cover
                predicate = None

        dataset = ds.dataset(
            input_catalog,
            format="parquet",
        )

        scan_node = acero.Declaration(
            "scan",
            acero.ScanNodeOptions(
                dataset,
                columns=COLUMNS_COM_CAM,
                filter=predicate,
            ),
        )

        filter_node = acero.Declaration(
            "filter",
            acero.FilterNodeOptions(
                predicate,
            ),
        )

        column_projection = {k: pc.field(k) for k in COLUMNS_COM_CAM}
        projection = column_projection
        project_nodes = []
        for _projection in PROJECTIONS_COM_CAM:
            for k, v in _projection.items():
                projection[k] = v
            project_node = acero.Declaration(
                "project",
                acero.ProjectNodeOptions(
                    [v for k, v in projection.items()],
                    names=[k for k, v in projection.items()],
                ),
            )
            project_nodes.append(project_node)

        seq = [
            scan_node,
            filter_node,
            *project_nodes,
        ]
        plan = acero.Declaration.from_sequence(seq)

        # batches = plan.to_reader(use_threads=True)
        table = plan.to_table(use_threads=True)
        print(f"writing dataset to {output_catalog}")

        output_dir = os.path.dirname(output_catalog)

        os.makedirs(output_dir, exist_ok=True)
        pd = table.to_pandas()
        pd.to_parquet(output_catalog)
        # pq.write_table(table, output_catalog)


class DP1Reducer(RailReducer):
    """Class to reduce the 'DP1' input files for pz analysis"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Reducer Name"),
        cuts=StageParameter(dict, {}, fmt="%s", msg="Selections"),
    )

    _mag_offset = 31.4

    def run(
        self,
        input_catalog: str,
        output_catalog: str,
    ) -> None:  # pragma: no cover

        topdir = os.path.dirname(os.path.dirname(input_catalog))
        columns_file = os.path.join(topdir, "columns.yaml")
        with open(columns_file, "r", encoding="utf-8") as fin:
            columns = yaml.safe_load(fin)

        # Try to do this right
        try:
            parsed_filter = parse_item(self.config.cuts)
            predicate = pq.filters_to_expression(parsed_filter)
        except Exception:
            # Fallback to old way.  FIXME, deprecate this
            mag_cut = self.config.cuts["maglim_i"][1]
            flux_cut = np.power(10, (self._mag_offset - mag_cut) / 2.5)
            if self.config.cuts:
                predicate = (
                    (pc.field("i_psfFlux") > flux_cut)
                    & (pc.field("i_psfFlux") / pc.field("i_psfFluxErr") > 5)
                    & (
                        (pc.field("g_extendedness") > 0.5)
                        | (pc.field("r_extendedness") > 0.5)
                    )
                )

            else:  # pragma: no cover
                predicate = None

        dataset = ds.dataset(
            input_catalog,
            format="parquet",
        )

        scan_node = acero.Declaration(
            "scan",
            acero.ScanNodeOptions(
                dataset,
                columns=columns,
                filter=predicate,
            ),
        )

        filter_node = acero.Declaration(
            "filter",
            acero.FilterNodeOptions(
                predicate,
            ),
        )

        column_projection = {k: pc.field(k) for k in columns}
        projection = column_projection
        project_nodes = []
        for _projection in PROJECTIONS_DP1:
            for k, v in _projection.items():
                projection[k] = v
            project_node = acero.Declaration(
                "project",
                acero.ProjectNodeOptions(
                    [v for k, v in projection.items()],
                    names=[k for k, v in projection.items()],
                ),
            )
            project_nodes.append(project_node)

        seq = [
            scan_node,
            filter_node,
            *project_nodes,
        ]
        plan = acero.Declaration.from_sequence(seq)

        # batches = plan.to_reader(use_threads=True)
        table = plan.to_table(use_threads=True)
        print(f"writing dataset to {output_catalog}")

        output_dir = os.path.dirname(output_catalog)

        os.makedirs(output_dir, exist_ok=True)
        pd = table.to_pandas()
        pd.to_parquet(output_catalog)
        # pq.write_table(table, output_catalog)



import os

from rail.projects import library


def test_libray_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    library.clear()

    library.load_yaml("tests/ci_project_library.yaml")
    library.print_contents()
    library.clear()

    library.write_yaml("tests/temp.yaml")
    library.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")
    library.clear()



from __future__ import annotations

from typing import Any

from ceci.config import StageParameter

from rail.projects import RailProject, path_funcs

from .data_extraction_funcs import (
    get_tomo_bins_nz_estimate_data,
    get_tomo_bins_true_nz_data,
)
from .dataset import RailDataset
from .dataset_factory import RailDatasetFactory
from .dataset_holder import (
    DatasetSplitMode,
    RailDatasetHolder,
    RailDatasetListHolder,
    RailProjectHolder,
)
from .nz_plotters import RailNZTomoBinsDataset


class RailNZTomoBinsDataHolder(RailDatasetHolder):
    """Simple class for holding a dataset for plotting data that comes from a RailProject"""

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Dataset name"),
        project=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject name"
        ),
        selection=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject data selection"
        ),
        flavor=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject analysis flavor"
        ),
        algo=StageParameter(
            str, None, fmt="%s", required=True, msg="RailProject algorithm"
        ),
        classifier=StageParameter(
            str, None, fmt="%s", required=True, msg="Tomographic bin classifier"
        ),
        summarizer=StageParameter(
            str, None, fmt="%s", required=True, msg="p(z) to n(z) summarizer"
        ),
    )

    extractor_inputs: dict = {
        "project": RailProject,
        "selection": str,
        "flavor": str,
        "algo": str,
        "classifier": str,
        "summarizer": str,
    }

    output_type: type[RailDataset] = RailNZTomoBinsDataset

    def __init__(self, **kwargs: Any):
        RailDatasetHolder.__init__(self, **kwargs)
        self._project: RailProject | None = None

    def __repr__(self) -> str:
        ret_str = (
            f"{self.__class__.__name__}"
            "( "
            f"{self.config.project}, "
            f"{self.config.selection}_{self.config.flavor}_{self.config.algo}_"
            f"{self.config.classifier}_{self.config.summarizer} "
            ")"
        )
        return ret_str

    def _get_data(self, **kwargs: Any) -> dict[str, Any]:
        kwcopy = kwargs.copy()
        kwcopy.pop("summarizer")
        data = dict(
            nz_estimates=get_tomo_bins_nz_estimate_data(**kwargs),
            truth=get_tomo_bins_true_nz_data(**kwcopy),
        )
        return data

    def get_extractor_inputs(self) -> dict[str, Any]:
        if self._project is None:
            self._project = RailDatasetFactory.get_project(
                self.config.project
            ).resolve()
        the_extractor_inputs = dict(
            project=self._project,
            selection=self.config.selection,
            flavor=self.config.flavor,
            algo=self.config.algo,
            classifier=self.config.classifier,
            summarizer=self.config.summarizer,
        )
        self._validate_extractor_inputs(**the_extractor_inputs)
        return the_extractor_inputs

    @classmethod
    def generate_dataset_dict(
        cls,
        **kwargs: Any,
    ) -> tuple[
        list[RailProjectHolder], list[RailDatasetHolder], list[RailDatasetListHolder]
    ]:
        """
        Parameters
        ----------
        **kwargs
            Set Notes

        Notes
        -----
        dataset_list_name: str
            Name for the resulting DatasetList

        project_file: str
            Config file for project to inspect

        selections: list[str]
            Selections to use

        flavors: list[str]
            Flavors to use

        classifiers: list[str]
            Flavors to use

        summarizers: list[str]
            Summarizers to use

        Returns
        -------
        list[RailProjectHolder]
            Underlying RailProjects

        list[RailDatasetHolder]
            Extracted datasets

        list[RailDatasetListHolder]
            Extracted dataset lists
        """
        dataset_list_name: str | None = kwargs.get("dataset_list_name")
        project_file = kwargs["project_file"]
        project = RailProject.load_config(project_file)

        selections = kwargs.get("selections")
        flavors = kwargs.get("flavors")
        split_mode = kwargs.get("split_mode", DatasetSplitMode.by_algo)

        flavor_dict = project.get_flavors()
        if flavors is None or "all" in flavors:
            flavors = list(flavor_dict.keys())
        if selections is None or "all" in selections:
            selections = list(project.get_selections().keys())

        project_name = project.name
        if not dataset_list_name:
            dataset_list_name = f"{project_name}_pz_point"

        projects: list[RailProjectHolder] = []
        datasets: list[RailDatasetHolder] = []
        dataset_lists: list[RailDatasetListHolder] = []

        projects.append(
            RailProjectHolder(
                name=project_name,
                yaml_file=project_file,
            )
        )

        dataset_list_dict: dict[str, list[str]] = {}
        dataset_key = dataset_list_name
        if split_mode == DatasetSplitMode.no_split:
            dataset_list_dict[dataset_key] = []

        for key in flavors:
            val = flavor_dict[key]
            pipelines = val["pipelines"]
            if "all" not in pipelines and "pz" not in pipelines:  # pragma: no cover
                continue
            try:
                algos = val["pipeline_overrides"]["default"]["kwargs"]["algorithms"]
            except KeyError:
                algos = list(project.get_pzalgorithms().keys())

            try:
                classifiers = val["pipeline_overrides"]["default"]["kwargs"][
                    "classifiers"
                ]
            except KeyError:
                classifiers = list(project.get_classifiers().keys())

            try:
                summarizers = val["pipeline_overrides"]["default"]["kwargs"][
                    "summarizers"
                ]
            except KeyError:
                summarizers = list(project.get_summarizers().keys())

            for selection_ in selections:
                if split_mode == DatasetSplitMode.by_flavor:
                    dataset_key = f"{dataset_list_name}_{selection_}_{key}"
                    dataset_list_dict[dataset_key] = []

                for algo_ in algos:
                    if split_mode == DatasetSplitMode.by_algo:
                        dataset_key = f"{dataset_list_name}_{selection_}_{algo_}"
                        if dataset_key not in dataset_list_dict:
                            dataset_list_dict[dataset_key] = []

                    for classifier_ in classifiers:
                        nz_true_paths = path_funcs.get_ceci_true_nz_output_paths(
                            project,
                            selection=selection_,
                            flavor=key,
                            algo=algo_,
                            classifier=classifier_,
                        )

                        if not nz_true_paths:  # pragma: no cover
                            continue

                        for summarizer_ in summarizers:
                            nz_paths = path_funcs.get_ceci_nz_output_paths(
                                project,
                                selection=selection_,
                                flavor=key,
                                algo=algo_,
                                classifier=classifier_,
                                summarizer=summarizer_,
                            )

                            if not nz_paths:  # pragma: no cover
                                continue

                            dataset_name = f"{selection_}_{key}_{algo_}_{classifier_}_{summarizer_}"

                            dataset = cls(
                                name=dataset_name,
                                project=project_name,
                                flavor=key,
                                algo=algo_,
                                selection=selection_,
                                classifier=classifier_,
                                summarizer=summarizer_,
                            )
                            datasets.append(dataset)
                            dataset_list_dict[dataset_key].append(dataset_name)

        for ds_name, ds_list in dataset_list_dict.items():
            # Skip empty lists
            if not ds_list:
                continue
            dataset_list = RailDatasetListHolder(
                name=ds_name,
                dataset_class=cls.output_type.full_class_name(),
                datasets=ds_list,
            )
            dataset_lists.append(dataset_list)

        return (projects, datasets, dataset_lists)



import os

from rail.plotting.dataset_holder import DatasetSplitMode
from rail.plotting.plot_group import RailPlotGroup
from rail.plotting.plot_group_factory import RailPlotGroupFactory


def test_load_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    RailPlotGroupFactory.clear()
    RailPlotGroupFactory.load_yaml("tests/ci_plot_groups.yaml")
    RailPlotGroupFactory.print_contents()

    # Make sure the names of the plot_groups got loaded
    the_plot_group_names = RailPlotGroupFactory.get_plot_group_names()
    assert "accuracy_v_ztrue" in the_plot_group_names

    # Make sure the plot_groups got loaded
    the_dict = RailPlotGroupFactory.get_plot_groups()
    assert isinstance(the_dict["accuracy_v_ztrue"], RailPlotGroup)

    the_plot_group = RailPlotGroupFactory.get_plot_group("accuracy_v_ztrue")
    assert isinstance(the_plot_group, RailPlotGroup)

    # Test the interactive stuff
    RailPlotGroupFactory.clear()

    RailPlotGroupFactory.add_plot_group(the_plot_group)

    check_plot_group = RailPlotGroupFactory.get_plot_group("accuracy_v_ztrue")
    assert isinstance(check_plot_group, RailPlotGroup)

    # check writing the yaml dict
    RailPlotGroupFactory.write_yaml("tests/temp.yaml")
    RailPlotGroupFactory.clear()
    RailPlotGroupFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_plot_group = RailPlotGroupFactory.get_plot_group("accuracy_v_ztrue")
    assert isinstance(check_plot_group, RailPlotGroup)


def test_make_yaml_for_dataset_list(setup_project_area: int) -> None:
    assert setup_project_area == 0

    RailPlotGroupFactory.make_yaml_for_dataset_list(
        output_yaml="tests/temp_data/check_pz_plot_groups.yaml",
        plotter_yaml_path="tests/ci_plots.yaml",
        dataset_yaml_path="tests/ci_datasets.yaml",
        plotter_list_name="zestimate_v_ztrue",
        output_prefix="",
        dataset_list_name="baseline_test",
    )
    RailPlotGroupFactory.make_yaml_for_dataset_list(
        output_yaml="tests/temp_data/check_nz_plot_groups.yaml",
        plotter_yaml_path="tests/ci_plots.yaml",
        dataset_yaml_path="tests/ci_datasets.yaml",
        plotter_list_name="tomo_bins",
        output_prefix="",
        dataset_list_name="blend_baseline_tomo_knn",
    )


def test_make_yaml_for_project(setup_project_area: int) -> None:
    assert setup_project_area == 0

    RailPlotGroupFactory.make_yaml_for_project(
        output_yaml="tests/temp_data/check_project_plot_groups_by_algo.yaml",
        plotter_yaml_path="tests/ci_plots.yaml",
        project_yaml_path="tests/ci_project.yaml",
        flavors=["all"],
        split_mode=DatasetSplitMode.by_algo,
    )
    RailPlotGroupFactory.make_yaml_for_project(
        output_yaml="tests/temp_data/check_project_plot_groups_by_flavor.yaml",
        plotter_yaml_path="tests/ci_plots.yaml",
        project_yaml_path="tests/ci_project.yaml",
        flavors=["all"],
        split_mode=DatasetSplitMode.by_flavor,
    )



from __future__ import annotations

from typing import Any

from ceci.config import StageParameter
from ceci.stage import PipelineStage
from rail.core.configurable import Configurable

from .dynamic_class import DynamicClass
from .reducer import RailReducer
from .splitter import RailSplitter
from .subsampler import RailSubsampler


class RailAlgorithmHolder(Configurable, DynamicClass):
    """Simple class for holding an algorithm by name.

    This has the information needed to create the associated classes,
    namely the name of the python module in which they live, and the
    names of the classes themselves.
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Algorithm name"),
        Module=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Name of associated module",
        ),
    )

    sub_classes: dict[str, type[DynamicClass]] = {}
    """Dictionary of sub_classes"""

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        kwargs: Any
            Configuration parameters for this RailAlgorithmHolder, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)
        DynamicClass.__init__(self)

    def __repr__(self) -> str:
        vals = self.config.to_dict().copy()
        vals.pop("name")
        vals.pop("Module")
        return f"{self.config.Module}.{vals}"

    def resolve(self, key: str) -> type:
        """Get the associated class one of the parts of the algorithm"""
        try:
            class_name = self.config[key]
        except KeyError as missing_key:
            raise KeyError(
                f"RailAlgorithmHolder does not have {key} in {self.config.to_dict().keys}"
            ) from missing_key
        return PipelineStage.get_stage(class_name, self.config.Module)

    def fill_dict(self, the_dict: dict[str, dict[str, str]]) -> None:
        """Fill a dict with infomation about the algorithm"""
        copy_dict = self.config.to_dict().copy()
        the_name = copy_dict.pop("name")
        the_dict[the_name] = copy_dict


class RailPZAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that estimate per-object p(z).

    This wraps both the Inform and Estimate classes.

    The Inform class will typically be a `CatInformer` type `RailStage`,
    used to train the model for p(z) estimation.

    The Estimate class will typically be a `CatEstimator` type `RailStage`,
    which uses the trained model for p(z) estimation.

    A set of PZAlgorithm are used as inputs to several of the pipelines,
    specifying that the set of algorithms to run the pipeline with.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Estimate=StageParameter(
            str, None, fmt="%s", required=True, msg="Estimator Class"
        ),
        Inform=StageParameter(str, None, fmt="%s", required=True, msg="Informer Class"),
    )
    yaml_tag = "PZAlgorithm"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)


class RailSummarizerAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that make ensemble n(z) from a set of p(z).

    This wraps the Summarize class,
    which is typically a `PZToNZSummarizer` type `RailStage`.

    A set of Summarizer are used as inputs to the tomography-related pipelines,
    specifying that the set of algorithms to obtain n(z) information.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Summarize=StageParameter(
            str, None, fmt="%s", required=True, msg="Summarizer Class"
        ),
    )
    yaml_tag = "Summarizer"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)


class RailClassificationAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that assign objects to tomographic bins.

    This wraps the Classify class,
    which is typically a `Classifier` type `RailStage`.

    A set of Classifier are used as inputs to the tomography-related pipelines,
    specifying that the set of algorithms to assign objects to tomographic bins.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Classify=StageParameter(
            str, None, fmt="%s", required=True, msg="Classifier Class"
        ),
    )
    yaml_tag = "Classifier"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)


class RailSpecSelectionAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that emulate spectrosopic selections.

    This wraps the SpecSelection class,
    which is typically a `SpecSelector` type `RailStage`.

    A set of SpecSelection are used as inputs to the observation emulation pipelines,
    specifying that the set of algorithms to emulate spectrosopic selections.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Select=StageParameter(str, None, fmt="%s", required=True, msg="Selector Class"),
    )
    yaml_tag = "SpecSelection"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)


class RailErrorModelAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that emulate photometric errors.

    This wraps the ErrorModel class,
    which is typically a `PhotoErrorModel` type RailStage

    A set of ErrorModel are used as inputs to the observation emulation pipelines,
    specifying that the set of algorithms to emulate photometric errors.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        ErrorModel=StageParameter(
            str, None, fmt="%s", required=True, msg="Photometric Error Model Class"
        ),
        Bands=StageParameter(
            list,
            ["u", "g", "r", "i", "z", "y"],
            fmt="%s",
            required=False,
            msg="Bands to apply errors to",
        ),
        Overrides=StageParameter(
            dict,
            {},
            fmt="%s",
            required=False,
            msg="Parameter overrides",
        ),
    )
    yaml_tag = "ErrorModel"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)


class RailReducerAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that reduce data sets by applying selections
    and removing unneed columns.

    This wraps the Reduce class, which is typically a `RailReducer` object.

    Typically a single Reducer is used to prepare data for a particular project,
    possible apply a few different selections along the way.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Reduce=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Data Reducer Class",
        ),
    )
    yaml_tag = "Reducer"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)

    def resolve(self, key: str) -> type:
        """Get the associated class one of the parts of the algorithm"""
        try:
            class_name = self.config[key]
        except KeyError as missing_key:
            raise KeyError(
                f"RailReducerAlgorithmHolder does not have {key} in {self.config.to_dict().keys}"
            ) from missing_key
        return RailReducer.get_sub_class(
            class_name, f"{self.config.Module}.{class_name}"
        )


class RailSubsamplerAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that sumsample catalogs
    to provide testing and training data sets.

    This wraps the Subsample class, which is typically a `RailSubsampler` object.

    Typically a single Subsample is used to create a number of different
    test and training data sets for a particular project.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Subsample=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Data Subsampler Class",
        ),
    )
    yaml_tag = "Subsampler"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)

    def resolve(self, key: str) -> type:
        """Get the associated class one of the parts of the algorithm"""
        try:
            class_name = self.config[key]
        except KeyError as missing_key:
            raise KeyError(
                f"RailSubsamplerAlgorithmHolder does not have {key} in {self.config.to_dict().keys}"
            ) from missing_key
        return RailSubsampler.get_sub_class(
            class_name, f"{self.config.Module}.{class_name}"
        )


class RailSplitterAlgorithmHolder(RailAlgorithmHolder):
    """Wrapper for algorithms that split files
    to provide testing and training data sets.

    This wraps the Splitter class, which is typically a `RailSplitter` object.

    Typically a single Splitter is used to create a number of different
    test and training data sets for a particular project.
    """

    config_options = RailAlgorithmHolder.config_options.copy()
    config_options.update(
        Split=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Data splitter Class",
        ),
    )
    yaml_tag = "Splitter"

    def __init__(self, **kwargs: Any):
        RailAlgorithmHolder.__init__(self, **kwargs)

    def resolve(self, key: str) -> type:
        """Get the associated class one of the parts of the algorithm"""
        try:
            class_name = self.config[key]
        except KeyError as missing_key:
            raise KeyError(
                f"RailSplitterAlgorithmHolder does not have {key} in {self.config.to_dict().keys}"
            ) from missing_key
        return RailSplitter.get_sub_class(
            class_name, f"{self.config.Module}.{class_name}"
        )



"""
The :py:mod:`rail.plotting` sub-package collects a set of tools to make
plots from the data generated by using the :py:mod:`rail.projects` tools.
"""

from . import control

__all__ = ["control"]
